import logging
import re
import warnings
from threading import Lock
from typing import Any

import networkx as nx
from transformers import pipeline

from nxlu.config import COMPLEXITY_COST_MAPPING, Intent, RescalingMethod
from nxlu.constants import REQUIRES_SOURCE_TARGET
from nxlu.processing.analyze import (
    GraphProperties,
    apply_algorithm,
    map_algorithm_result,
)
from nxlu.processing.preprocess import (
    CleanGraphConfig,
    GraphRescalingConfig,
    GraphThresholdingConfig,
    SubgraphSelectionConfig,
)
from nxlu.utils.control import ResourceManager

warnings.filterwarnings("ignore")


logger = logging.getLogger("nxlu")


__all__ = [
    "AlgorithmNominator",
    "GraphAlgorithmElector",
    "GraphPreprocessingSelector",
    "apply_algorithm",
    "map_algorithm_result",
    "GraphProperties",
    "GraphRescalingConfig",
    "GraphThresholdingConfig",
    "SubgraphSelectionConfig",
    "CleanGraphConfig",
    "ResourceManager",
    "COMPLEXITY_COST_MAPPING",
    "RescalingMethod",
]


class AlgorithmNominator:
    """Select applicable graph algorithms based on graph properties and resource
    constraints.
    """

    def __init__(
        self,
        applicability_dict: dict[str, dict[str, Any]],
        resource_manager: ResourceManager,
        include_algorithms: list[str] | None = None,
        exclude_algorithms: list[str] | None = None,
        enable_classification: bool = True,
    ):
        """Initialize the AlgorithmNominator.

        Parameters
        ----------
        applicability_dict : Dict[str, Dict[str, Any]]
            Dictionary containing algorithm applicability conditions.
        resource_manager : ResourceManager
            Manager to handle resource constraints.
        """
        self.applicability_dict = applicability_dict
        self.resource_manager = resource_manager
        self.include_algorithms = (
            set(include_algorithms) if include_algorithms else set()
        )
        self.exclude_algorithms = (
            set(exclude_algorithms) if exclude_algorithms else set()
        )
        self.enable_classification = enable_classification
        self._algorithms_lock = Lock()

    def select_algorithms(
        self, graph_properties: GraphProperties, query: str
    ) -> list[str]:
        """Select algorithms applicable to the given graph properties.

        Parameters
        ----------
        graph_properties : GraphProperties
            Precomputed properties of the graph.
        query : str
            The user's query.

        Returns
        -------
        List[str]
            List of selected algorithm names.
        """
        if not self.enable_classification:
            elected = [
                alg
                for alg in self.include_algorithms
                if alg not in self.exclude_algorithms and alg in self.applicability_dict
            ]
            logger.info(
                f"Classification disabled. Selected algorithms based on inclusion and "
                f"exclusion: {elected}"
            )
            return elected

        with self._algorithms_lock:
            num_nodes = graph_properties.num_nodes
            num_edges = graph_properties.num_edges
            selected_algorithms = []

            logger.info(
                f"Selecting applicable algorithms for graph with {num_nodes} nodes "
                f"and {num_edges} edges."
            )

            for alg, conditions in self.applicability_dict.items():
                logger.debug(f"Evaluating applicability for algorithm: '{alg}'")
                if self._is_applicable(graph_properties, conditions):
                    time_complexity = conditions.get(
                        "time_complexity", "unknown"
                    ).lower()
                    space_complexity = conditions.get(
                        "space_complexity", "unknown"
                    ).lower()

                    time_cost_func = COMPLEXITY_COST_MAPPING.get(time_complexity)
                    space_cost_func = COMPLEXITY_COST_MAPPING.get(space_complexity)

                    if not time_cost_func or not space_cost_func:
                        logger.warning(
                            f"Algorithm '{alg}' has unknown complexity types "
                            f"(Time: '{time_complexity}', Space: '{space_complexity}')."
                            f"Excluding from selection."
                        )
                        continue

                    estimated_time = time_cost_func(num_nodes, num_edges)
                    estimated_space = space_cost_func(num_nodes, num_edges)

                    logger.debug(
                        f"Estimated costs for '{alg}': Time = {estimated_time}, "
                        f"Space = "
                        f"{estimated_space}"
                    )

                    if self.resource_manager.is_within_limits(
                        estimated_time, estimated_space
                    ):
                        selected_algorithms.append(alg)
                        logger.info(
                            f"Algorithm '{alg}' is selected (Estimated Time: "
                            f"{estimated_time}, "
                            f"Estimated Space: {estimated_space})."
                        )
                    else:
                        logger.warning(
                            f"Algorithm '{alg}' excluded due to resource constraints "
                            f"(Estimated Time: {estimated_time}, Estimated Space: "
                            f"{estimated_space})."
                        )
                else:
                    logger.debug(
                        f"Algorithm '{alg}' is not applicable based on graph "
                        f"properties."
                    )

            if self.exclude_algorithms:
                selected_algorithms = [
                    alg
                    for alg in selected_algorithms
                    if alg not in self.exclude_algorithms
                ]
                logger.info(f"Algorithms after exclusion: {selected_algorithms}")

            if self.include_algorithms:
                included = [
                    alg
                    for alg in self.include_algorithms
                    if alg in self.applicability_dict
                ]
                selected_algorithms.extend(
                    [alg for alg in included if alg not in selected_algorithms]
                )
                logger.info(f"Algorithms after inclusion: {selected_algorithms}")

            selected_algorithms = self._filter_path_algorithms(
                selected_algorithms, query
            )

            logger.info(f"Selected algorithms: {selected_algorithms}")
            return selected_algorithms

    @staticmethod
    def _filter_path_algorithms(algorithms: list[str], query: str) -> list[str]:
        """Filter out algorithms that require source/target nodes if not applicable.

        Parameters
        ----------
        algorithms : list of str
            The list of algorithm names to filter.
        query : str
            The user's query.

        Returns
        -------
        list of str
            The filtered list of algorithm names.
        """
        has_source = bool(re.search(r"from\s+(['\"]?)(\w+)\1", query, re.IGNORECASE))
        has_target = bool(re.search(r"to\s+(['\"]?)(\w+)\1", query, re.IGNORECASE))

        filtered_algorithms = []
        for algo in algorithms:
            if algo in REQUIRES_SOURCE_TARGET:
                if has_source and (
                    algo
                    not in {
                        "single_source_shortest_path",
                        "single_source_dijkstra_path",
                        "single_source_bellman_ford_path",
                    }
                    or has_target
                ):
                    filtered_algorithms.append(algo)
                else:
                    logger.info(
                        f"Excluding algorithm '{algo}' due to missing source/target."
                    )
            else:
                filtered_algorithms.append(algo)

        return filtered_algorithms

    def _is_applicable(
        self, graph_props: GraphProperties, conditions: dict[str, Any]
    ) -> bool:
        """Determine if an algorithm is applicable to the given graph based on its
        conditions.

        Parameters
        ----------
        graph_props : GraphProperties
            Precomputed properties of the graph.
        conditions : Dict[str, Any]
            Applicability conditions for the algorithm.

        Returns
        -------
        bool
            True if applicable, False otherwise.
        """
        # Directedness
        if "directed" in conditions:
            if conditions["directed"] and not graph_props.is_directed:
                logger.debug(
                    "Algorithm requires directed graph, but graph is undirected."
                )
                return False
            if not conditions["directed"] and graph_props.is_directed:
                logger.debug(
                    "Algorithm requires undirected graph, but graph is directed."
                )
                return False

        # Connectedness
        if "connected" in conditions:
            if conditions["connected"] and not graph_props.is_connected:
                logger.debug(
                    "Algorithm requires connected graph, but graph is disconnected."
                )
                return False

        # Strongly Connectedness
        if "strongly_connected" in conditions:
            if (
                conditions["strongly_connected"]
                and not graph_props.is_strongly_connected
            ):
                logger.debug(
                    "Algorithm requires strongly connected graph, but it is not."
                )
                return False

        # Weakly Connectedness
        if "weakly_connected" in conditions:
            if conditions["weakly_connected"] and not graph_props.is_connected:
                logger.debug(
                    "Algorithm requires weakly connected graph, but it is not."
                )
                return False

        # Bipartite
        if "bipartite" in conditions:
            if conditions["bipartite"] and not graph_props.is_bipartite:
                logger.debug("Algorithm requires bipartite graph, but it is not.")
                return False

        # Multigraph
        if "multigraph" in conditions:
            if conditions["multigraph"] and not graph_props.is_multigraph:
                logger.debug("Algorithm requires multigraph, but graph is not.")
                return False
            if not conditions["multigraph"] and graph_props.is_multigraph:
                logger.debug(
                    "Algorithm requires simple graph, but graph is a multigraph."
                )
                return False

        # Simple Graph
        if "simple_graph" in conditions:
            if conditions["simple_graph"] and graph_props.is_multigraph:
                logger.debug(
                    "Algorithm requires simple graph, but graph is a multigraph."
                )
                return False

        # Node and Edge Counts
        if (
            "min_nodes" in conditions
            and graph_props.num_nodes < conditions["min_nodes"]
        ):
            logger.debug(
                f"Algorithm requires at least {conditions['min_nodes']} nodes, "
                f"but graph has {graph_props.num_nodes}."
            )
            return False
        if (
            "max_nodes" in conditions
            and graph_props.num_nodes > conditions["max_nodes"]
        ):
            logger.debug(
                f"Algorithm requires at most {conditions['max_nodes']} nodes, "
                f"but graph has {graph_props.num_nodes}."
            )
            return False
        if (
            "min_edges" in conditions
            and graph_props.num_edges < conditions["min_edges"]
        ):
            logger.debug(
                f"Algorithm requires at least {conditions['min_edges']} edges, "
                f"but graph has {graph_props.num_edges}."
            )
            return False
        if (
            "max_edges" in conditions
            and graph_props.num_edges > conditions["max_edges"]
        ):
            logger.debug(
                f"Algorithm requires at most {conditions['max_edges']} edges, "
                f"but graph has {graph_props.num_edges}."
            )
            return False

        # Data Presence
        if (
            "has_edge_data" in conditions
            and conditions["has_edge_data"]
            and not graph_props.has_edge_data
        ):
            logger.debug("Algorithm requires edge data, but graph edges lack data.")
            return False
        if (
            "has_node_data" in conditions
            and conditions["has_node_data"]
            and not graph_props.has_node_data
        ):
            logger.debug("Algorithm requires node data, but graph nodes lack data.")
            return False

        # Density
        if "density" in conditions:
            density_condition = conditions["density"]
            if isinstance(density_condition, str):
                if density_condition == "dense" and graph_props.density < 0.5:
                    logger.debug(
                        f"Algorithm requires dense graph, but density is "
                        f"{graph_props.density:.2f}."
                    )
                    return False
                if density_condition == "sparse" and graph_props.density >= 0.5:
                    logger.debug(
                        f"Algorithm requires sparse graph, but density is "
                        f"{graph_props.density:.2f}."
                    )
                    return False
            elif isinstance(density_condition, tuple):
                min_density, max_density = density_condition
                if not (min_density <= graph_props.density <= max_density):
                    logger.debug(
                        f"Algorithm requires density between {min_density} and "
                        f"{max_density}, "
                        f"but density is {graph_props.density:.2f}."
                    )
                    return False

        # Planarity
        if (
            "planar" in conditions
            and conditions["planar"]
            and not graph_props.is_planar
        ):
            logger.debug("Algorithm requires planar graph, but graph is not planar.")
            return False

        # Tree Structure
        if (
            "is_tree" in conditions
            and conditions["is_tree"]
            and not graph_props.is_tree
        ):
            logger.debug("Algorithm requires tree structure, but graph is not a tree.")
            return False

        # Cycles
        if "has_cycles" in conditions:
            has_cycles = not nx.is_forest(graph_props.graph)
            if conditions["has_cycles"] and not has_cycles:
                logger.debug("Algorithm requires cycles, but graph has none.")
                return False
            if not conditions["has_cycles"] and has_cycles:
                logger.debug("Algorithm requires acyclic graph, but graph has cycles.")
                return False

        logger.debug("Algorithm is applicable based on graph properties.")
        return True


class GraphAlgorithmElector:
    """Elect and apply graph algorithms based on user queries and graph summaries."""

    def __init__(
        self,
        algorithm_docs: dict[str, str],
        applicability_dict: dict[str, dict[str, Any]],
        include_algorithms: list[str] | None = None,
        exclude_algorithms: list[str] | None = None,
        enable_classification: bool = True,
    ):
        """Initialize the GraphAlgorithmElector.

        Parameters
        ----------
        algorithm_docs : Dict[str, str]
            Dictionary containing documentation strings for algorithms.
        applicability_dict : Dict[str, Dict[str, Any]]
            Dictionary containing algorithm applicability conditions.
        """
        self.zero_shot_classifier = pipeline(
            "zero-shot-classification", model="facebook/bart-large-mnli", device=-1
        )
        self.algorithm_docs = algorithm_docs
        self.applicability_dict = applicability_dict
        self.include_algorithms = (
            set(include_algorithms) if include_algorithms else set()
        )
        self.exclude_algorithms = (
            set(exclude_algorithms) if exclude_algorithms else set()
        )
        self.enable_classification = enable_classification

    def elect_algorithms(
        self,
        query: str,
        graph_summary: str,
        user_intent: list[Intent],
        candidates: list[str],
        top_n: int = 5,
    ) -> list[str]:
        """Select algorithms using zero-shot classification based on query and graph
        summary.

        Parameters
        ----------
        query : str
            The user's query.
        graph_summary : str
            Summary of the graph structure and properties.
        user_intent : List[Intent]
            List of high-level intents detected.
        candidates : List[str]
            List of applicable algorithms to choose from.
        top_n : int, optional
            Number of top algorithms to select, by default 5.

        Returns
        -------
        List[str]
            List of selected algorithm names.
        """
        if not self.enable_classification:
            elected = [alg for alg in self.include_algorithms if alg in candidates]
            logger.info(
                f"Classification disabled. Elected algorithms based on inclusion: "
                f"{elected}"
            )
            return elected

        intent_descriptions = [intent.value for intent in user_intent]

        candidate_descriptions = []
        for alg in candidates:
            doc = self.algorithm_docs.get(alg, {})
            technical = doc.get("technical", "")
            colloquial = doc.get("colloquial", "")
            description = f"{technical} {colloquial}".strip()
            candidate_descriptions.append(description)

        candidate_descriptions = [
            self.algorithm_docs.get(alg, "") for alg in candidates
        ]

        if not candidate_descriptions:
            raise ValueError("No candidate algorithm descriptions found.")

        prompt_template = """
            Given the user query: '{query}', the graph summary: '{graph_summary}',
            and the user's intents: '{intents}', which of the networkx algorithms would
            generate the most relevant insights for responding helpfully to the user
            query?
        """

        # Escape user input to prevent any injection issues
        escaped_query = re.escape(query)
        escaped_graph_summary = re.escape(graph_summary)
        escaped_intents = ", ".join(map(re.escape, intent_descriptions))

        prompt = prompt_template.format(
            query=escaped_query,
            graph_summary=escaped_graph_summary,
            intents=escaped_intents,
        )

        try:
            result = self.zero_shot_classifier(
                prompt, candidate_descriptions, multi_label=True
            )
        except Exception:
            raise RuntimeError("Error during zero-shot classification")

        algorithm_scores = list(zip(candidates, result["scores"]))
        sorted_algorithms = sorted(algorithm_scores, key=lambda x: x[1], reverse=True)
        top_algorithms = sorted_algorithms[:top_n]

        elected_algorithms = [alg for alg, _ in top_algorithms]

        logger.info(f"Classified top algorithms: {elected_algorithms}")

        if self.exclude_algorithms:
            elected_algorithms = [
                alg for alg in elected_algorithms if alg not in self.exclude_algorithms
            ]
            logger.info(f"Algorithms after exclusion: {elected_algorithms}")

        if self.include_algorithms:
            for alg in self.include_algorithms:
                if alg not in elected_algorithms and alg in candidates:
                    elected_algorithms.append(alg)
                    logger.info(f"Algorithm '{alg}' included as per user preference.")

        logger.info(f"Final elected algorithms: {elected_algorithms}")
        return elected_algorithms

    def apply_elected_algorithms(
        self, graph: nx.Graph, algorithms: list[str], query: str
    ) -> dict[str, Any]:
        """Apply elected algorithms to the graph and map their results.

        Parameters
        ----------
        graph : nx.Graph
            The graph to analyze.
        algorithms : List[str]
            List of algorithms to apply.
        query : str
            The user's query.

        Returns
        -------
        Dict[str, Any]
            Dictionary of algorithm results.
        """
        results = {}

        def apply_algorithm_safe(algorithm):
            """Apply a single algorithm and handle exceptions."""
            try:
                kwargs = self._extract_algorithm_parameters(query, algorithm)
                result = apply_algorithm(
                    self.applicability_dict, graph, algorithm, **kwargs
                )
            except Exception:
                error_message = f"Error applying algorithm '{algorithm}'"
                logger.exception(error_message)
                return error_message
            else:
                map_algorithm_result(graph, algorithm, result)
                logger.info(f"Applied algorithm '{algorithm}' successfully.")
                return result

        for algorithm in algorithms:
            results[algorithm] = apply_algorithm_safe(algorithm)
        return results

    @staticmethod
    def _extract_algorithm_parameters(
        query: str, algorithm_name: str
    ) -> dict[str, Any]:
        """Extract additional parameters for the algorithm based on the user's query.

        Parameters
        ----------
        query : str
            The user's query.
        algorithm_name : str
            The name of the algorithm.

        Returns
        -------
        Dict[str, Any]
            Dictionary of parameters.
        """
        parameters = {}
        if not query:
            query = ""

        patterns = {
            "path": r"from\s+(['\"]?)(\w+)\1\s+to\s+(['\"]?)(\w+)\3",
            "source": r"from\s+(['\"]?)(\w+)\1",
            "target": r"to\s+(['\"]?)(\w+)\1",
        }

        def extract(pattern):
            match = re.search(pattern, query, re.IGNORECASE)
            return match.group(2) if match else None

        if algorithm_name in {
            "shortest_path",
            "dijkstra_path",
            "bellman_ford_path",
            "shortest_simple_paths",
        }:
            source = extract(patterns["path"]) or extract(patterns["source"])
            target = extract(patterns["path"]) or extract(patterns["target"])
            if source and target:
                parameters["source"] = source
                parameters["target"] = target
            else:
                logger.warning(
                    f"Source and/or target nodes not specified for {algorithm_name}. "
                    f"Skipping algorithm."
                )
                parameters["source"] = None
                parameters["target"] = None

        elif algorithm_name in {
            "single_source_shortest_path",
            "single_source_dijkstra_path",
            "single_source_bellman_ford_path",
        }:
            source = extract(patterns["source"])
            if source:
                parameters["source"] = source
            else:
                logger.warning(
                    f"Source node not specified for {algorithm_name}. Skipping "
                    f"algorithm."
                )
                parameters["source"] = None

        return parameters


class GraphPreprocessingSelector:
    """Predicts which preprocessing steps to apply based on the graph and the
    algorithms.
    """

    def __init__(self, algorithm_applicability: dict[str, dict[str, Any]]):
        """Initialize the GraphPreprocessingSelector.

        Parameters
        ----------
        algorithm_applicability : Dict[str, Dict[str, Any]]
            Dictionary containing algorithm applicability conditions.
        """
        self.algorithm_applicability = algorithm_applicability

    def select_preprocessing_steps(
        self, graph: nx.Graph, algorithms: list[str]
    ) -> CleanGraphConfig:
        """Determine the necessary preprocessing steps based on the graph and selected
        algorithms.

        Parameters
        ----------
        graph : nx.Graph
            The input graph.
        algorithms : List[str]
            List of algorithms to be applied.

        Returns
        -------
        CleanGraphConfig
            Configuration object specifying preprocessing steps.
        """
        logger.info("Predicting preprocessing steps based on selected algorithms.")

        config = CleanGraphConfig(
            force_symmetry=False,
            remove_self_loops=False,
            threshold=None,
            subgraph=None,
            rescale=None,
        )

        requires_symmetry = False
        requires_self_loops_removal = False
        requires_rescaling = False
        requires_connectedness = False
        rescaling_methods = set()

        for alg in algorithms:
            conditions = self.algorithm_applicability.get(alg, {})
            if not conditions:
                logger.warning(
                    f"No applicable conditions found for algorithm '{alg}'. Skipping."
                )
                continue

            if conditions.get("requires_symmetry", False):
                requires_symmetry = True

            if not conditions.get("self_loops", True):
                requires_self_loops_removal = True

            if conditions.get("weighted", False):
                requires_rescaling = True
                rescaling_methods.add(RescalingMethod.normalize)

            if conditions.get("requires_connectedness", False):
                requires_connectedness = True

        # apply aggregated preprocessing configurations
        # 1. Force Symmetry
        if requires_symmetry:
            config.force_symmetry = True
            logger.info("Configured to force symmetry in the graph.")

        # 2. Remove Self-Loops
        if requires_self_loops_removal or nx.number_of_selfloops(graph) > 0:
            config.remove_self_loops = True
            logger.info("Configured to remove self-loops from the graph.")

        # 3. Rescale Graph
        if requires_rescaling:
            if rescaling_methods:
                selected_rescaling = rescaling_methods.pop()
                config.rescale = GraphRescalingConfig(method=selected_rescaling)
                logger.info(
                    f"Configured to rescale the graph using method: "
                    f"{selected_rescaling.value}."
                )

        # 4. Apply Merged Subgraph Selection Configuration
        if requires_connectedness:
            config.subgraph = self._select_subgraph_method(
                graph, SubgraphSelectionConfig()
            )
            logger.info(f"Subgraph selection: {config.subgraph}")

        logger.info(f"Final preprocessing configuration: {config}")
        return config

    @staticmethod
    def _select_subgraph_method(
        G: nx.Graph, config: SubgraphSelectionConfig
    ) -> SubgraphSelectionConfig:
        """Determine which subgraph selection method to apply."""
        num_components = nx.number_connected_components(G)

        # more than 1 connected comp
        if num_components > 1:
            if config.min_nodes:
                # small components that need pruning
                small_components = [
                    comp
                    for comp in nx.connected_components(G)
                    if len(comp) < config.min_nodes
                ]
                if len(small_components) > 0:
                    config.prune_components = True
                    logger.debug("Merged 'prune_components' flag set to True.")
                    return config

            logger.debug("Merged 'use_lcc' flag set to True.")
            config.use_lcc = True
            return config

        # isolates
        isolates = [n for n, d in G.degree() if d == 0]
        if len(isolates) > 0 and config.defragment:
            config.defragment = True
            logger.debug("Merged 'defragment' flag set to True.")
            return config

        return config



import os
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any

from networkx.utils.configs import Config

from _nxlu.enums import AnthropicModel, Framework, LocalModel, OpenAIModel
from _nxlu.logging import LoggerConfig, LoggingConfig, LoggingHandlerConfig

__all__ = ["_config", "NxluConfig"]


@dataclass
class NxluConfig(Config):
    """Configuration for NetworkX that controls behaviors such as how to use backends
    and logging.
    """

    active: bool = False
    simple: int = False
    cache_converted_graphs: bool = True
    llm_framework: Framework | str = "langchain"
    openai_api_key: str | None = None
    anthropic_api_key: str | None = None
    model_name: OpenAIModel | AnthropicModel | LocalModel | str = "gpt-4o-mini"
    temperature: float = 0.1
    max_tokens: int = 5000
    num_thread: int = 8
    num_gpu: int = 0
    backend_name: str = "nxlu"
    backend_params: dict = field(default_factory=dict)

    logging_config: LoggingConfig = field(default_factory=LoggingConfig)

    include_algorithms: list[str] | None = field(default_factory=list)
    exclude_algorithms: list[str] | None = field(default_factory=list)
    enable_classification: bool = True

    _LLM_FRAMEWORK = "LLM_FRAMEWORK"
    _OPENAI_API_KEY = "OPENAI_API_KEY"
    _ANTHROPIC_API_KEY = "ANTHROPIC_API_KEY"
    _MODEL_NAME = "MODEL_NAME"
    _TEMPERATURE = "TEMPERATURE"
    _MAX_TOKENS = "MAX_TOKENS"
    _NUM_THREAD = "NUM_THREAD"
    _NUM_GPU = "NUM_GPU"
    _BACKEND_NAME = "BACKEND_NAME"

    _observers: list[Callable[[str, Any], None]] = field(
        default_factory=list, init=False, repr=False
    )

    def __post_init__(self):
        """Set environment variables based on initialized fields."""
        if self.llm_framework:
            self.set_llm_framework(self.llm_framework)
        if self.openai_api_key:
            self.set_openai_api_key(self.openai_api_key)
        if self.anthropic_api_key:
            self.set_anthropic_api_key(self.anthropic_api_key)
        if self.model_name:
            self.set_model_name(self.model_name)
        self.set_temperature(self.temperature)
        self.set_max_tokens(self.max_tokens)
        self.set_num_thread(self.num_thread)
        self.set_num_gpu(self.num_gpu)
        self.set_backend_name(self.backend_name)

    def register_observer(self, callback: Callable[[str, Any], None]) -> None:
        """Register an observer callback to be notified on configuration changes.

        Parameters
        ----------
        callback : Callable[[str, any], None]
            A function that accepts two arguments: the name of the configuration
            parameter
            that changed and its new value.
        """
        self._observers.append(callback)

    def notify_observers(self, name: str, value: Any) -> None:
        """Notify all registered observers about a configuration change.

        Parameters
        ----------
        name : str
            The name of the configuration parameter that changed.
        value : Any
            The new value of the configuration parameter.
        """
        for callback in self._observers:
            callback(name, value)

    def set_llm_framework(self, llm_framework: Framework | str) -> None:
        """Set the LLM framework, supporting both enum and string."""
        if isinstance(llm_framework, Framework):
            os.environ[self._LLM_FRAMEWORK] = llm_framework.value
        else:
            os.environ[self._LLM_FRAMEWORK] = llm_framework

    def get_llm_framework(self) -> str | None:
        """Get the LLM framework."""
        return os.environ.get(self._LLM_FRAMEWORK)

    def set_openai_api_key(self, key: str) -> None:
        """Set the OpenAI API key."""
        os.environ[self._OPENAI_API_KEY] = key

    def get_openai_api_key(self) -> str | None:
        """Get the OpenAI API key."""
        return os.environ.get(self._OPENAI_API_KEY)

    def set_anthropic_api_key(self, key: str) -> None:
        """Set the Anthropic API key."""
        os.environ[self._ANTHROPIC_API_KEY] = key

    def get_anthropic_api_key(self) -> str | None:
        """Get the Anthropic API key."""
        return os.environ.get(self._ANTHROPIC_API_KEY)

    def set_model_name(
        self, model: OpenAIModel | AnthropicModel | LocalModel | str
    ) -> None:
        """Set the model name, supporting both enum and string."""
        if isinstance(model, (OpenAIModel, AnthropicModel, LocalModel)):
            os.environ[self._MODEL_NAME] = model.value
        else:
            os.environ[self._MODEL_NAME] = model

    def get_model_name(self) -> str:
        """Get the model name."""
        return os.environ.get(self._MODEL_NAME, self.model_name)

    def set_temperature(self, temp: float) -> None:
        """Set the temperature."""
        os.environ[self._TEMPERATURE] = str(temp)

    def get_temperature(self) -> float:
        """Get the temperature."""
        return float(os.environ.get(self._TEMPERATURE, "0.2"))

    def set_max_tokens(self, tokens: int) -> None:
        """Set the maximum number of tokens."""
        os.environ[self._MAX_TOKENS] = str(tokens)

    def get_max_tokens(self) -> int:
        """Get the maximum number of tokens."""
        return int(os.environ.get(self._MAX_TOKENS, "1000"))

    def set_num_thread(self, threads: int) -> None:
        """Set the number of threads."""
        os.environ[self._NUM_THREAD] = str(threads)

    def get_num_thread(self) -> int:
        """Get the number of threads."""
        return int(os.environ.get(self._NUM_THREAD, "8"))

    def set_num_gpu(self, gpus: int) -> None:
        """Set the number of GPUs."""
        os.environ[self._NUM_GPU] = str(gpus)

    def get_num_gpu(self) -> int:
        """Get the number of GPUs."""
        return int(os.environ.get(self._NUM_GPU, "0"))

    def set_backend_name(self, name: str) -> None:
        """Set the backend name."""
        os.environ[self._BACKEND_NAME] = name

    def get_backend_name(self) -> str:
        """Get the backend name."""
        return os.environ.get(self._BACKEND_NAME, "nxlu")

    def set_verbosity_level(self, level: int) -> None:
        """Set the verbosity level (0-2). 2=DEBUG, 1=INFO, 0=NO logging."""
        if level not in [0, 1, 2]:
            raise ValueError("Verbosity level must be 0, 1, or 2")

        level_map = {0: None, 1: "INFO", 2: "DEBUG"}
        log_level = level_map[level]

        nxlu_logger_cfg = next(
            (logger for logger in self.logging_config.loggers if logger.name == "nxlu"),
            None,
        )

        if level == 0:
            if nxlu_logger_cfg:
                self.logging_config.loggers.remove(nxlu_logger_cfg)
                self.notify_observers("remove_logger", "nxlu")
        elif nxlu_logger_cfg:
            nxlu_logger_cfg.level = log_level
            for handler in nxlu_logger_cfg.handlers:
                handler.level = log_level
            self.notify_observers("update_logger", "nxlu")
        else:
            new_logger = LoggerConfig(
                name="nxlu",
                level=log_level,
                handlers=[
                    LoggingHandlerConfig(
                        handler_type="console",
                        level=log_level,
                        formatter="%(asctime)s - %(name)s - %(levelname)s - % "
                        "(message)s",
                    )
                ],
            )
            self.logging_config.loggers.append(new_logger)
            self.notify_observers("add_logger", "nxlu")

        self.notify_observers("verbosity_level", level)

    def set_include_algorithms(self, algorithms: list[str]) -> None:
        """Specify algorithms to include explicitly."""
        self.include_algorithms = algorithms
        self.notify_observers("include_algorithms", algorithms)

    def set_exclude_algorithms(self, algorithms: list[str]) -> None:
        """Specify algorithms to exclude explicitly."""
        self.exclude_algorithms = algorithms
        self.notify_observers("exclude_algorithms", algorithms)

    def set_enable_classification(self, enable: bool) -> None:
        """Enable or disable automatic algorithm classification."""
        self.enable_classification = enable
        self.notify_observers("enable_classification", enable)


_config = NxluConfig()



from unittest.mock import Mock

import networkx as nx
import psutil
import pytest

from nxlu.utils.control import NetworkXGraphStore, ResourceManager

# ==============================
# Fixtures for ResourceManager
# ==============================


@pytest.fixture
def mock_cpu_count(monkeypatch):
    """Fixture to mock psutil.cpu_count."""
    mock = Mock(return_value=4)
    monkeypatch.setattr(psutil, "cpu_count", mock)
    return mock


@pytest.fixture
def mock_cpu_freq(monkeypatch):
    """Fixture to mock psutil.cpu_freq."""
    mock = Mock(return_value=Mock(current=3.0))
    monkeypatch.setattr(psutil, "cpu_freq", mock)
    return mock


@pytest.fixture
def mock_virtual_memory(monkeypatch):
    """Fixture to mock psutil.virtual_memory."""
    mock = Mock(return_value=Mock(available=16 * 1024**3))  # 16 GB available
    monkeypatch.setattr(psutil, "virtual_memory", mock)
    return mock


@pytest.fixture
def resource_manager_default(mock_cpu_count, mock_cpu_freq, mock_virtual_memory):
    """Fixture for ResourceManager with default parameters."""
    return ResourceManager()


@pytest.fixture
def resource_manager_custom():
    """Fixture for ResourceManager with custom limits."""
    return ResourceManager(max_time=1000.0, max_space=8 * 1024**3)  # 8 GB


# ==============================
# Fixtures for NetworkXGraphStore
# ==============================


@pytest.fixture
def simple_graph():
    """Fixture for a simple NetworkX Graph."""
    G = nx.Graph()
    G.add_node("A", data="Alpha")
    G.add_node("B", data="Beta")
    G.add_edge("A", "B", relation="connected")
    return G


@pytest.fixture
def multi_graph():
    """Fixture for a NetworkX MultiGraph."""
    MG = nx.MultiGraph()
    MG.add_node("A", data="Alpha")
    MG.add_node("B", data="Beta")
    MG.add_edge("A", "B", relation="connected", key=0)
    MG.add_edge("A", "B", relation="linked", key=1)
    return MG


@pytest.fixture
def graph_store_simple(simple_graph):
    """Fixture for NetworkXGraphStore with a simple graph."""
    return NetworkXGraphStore(graph=simple_graph)


@pytest.fixture
def graph_store_multi(multi_graph):
    """Fixture for NetworkXGraphStore with a multigraph."""
    return NetworkXGraphStore(graph=multi_graph)


# ==============================
# Tests for ResourceManager
# ==============================


class TestResourceManagerInitialization:
    def test_default_initialization(self, resource_manager_default):
        """Test ResourceManager initializes with dynamic limits when None provided."""
        assert resource_manager_default.max_time > 0
        assert resource_manager_default.max_space > 0

    def test_custom_initialization(self, resource_manager_custom):
        """Test ResourceManager initializes with provided limits."""
        assert resource_manager_custom.max_time == 1000.0
        assert resource_manager_custom.max_space == 8 * 1024**3

    def test_dynamic_max_time_calculation(
        self, monkeypatch, mock_cpu_count, mock_cpu_freq, mock_virtual_memory
    ):
        """Test dynamic calculation of max_time based on CPU resources."""
        monkeypatch.setattr(psutil, "cpu_count", Mock(return_value=8))
        monkeypatch.setattr(psutil, "cpu_freq", Mock(return_value=Mock(current=2.5)))
        rm = ResourceManager()
        expected_max_time = min(8 * 30 * (2.5 / 2.5), 3600)  # 240.0
        assert rm.max_time == expected_max_time

    def test_dynamic_max_space_calculation(self, mock_virtual_memory):
        """Test dynamic calculation of max_space based on available memory."""
        rm = ResourceManager()
        expected_max_space = 16 * 1024**3 * 0.9  # 90% of 16 GB
        assert rm.max_space == expected_max_space


class TestResourceManagerIsWithinLimits:
    @pytest.mark.parametrize(
        ("estimated_time", "estimated_space", "expected"),
        [
            (500.0, 8 * 1024**3, True),
            (1000.0, 8 * 1024**3, True),
            (1001.0, 8 * 1024**3, False),
            (500.0, 8 * 1024**3 + 1, False),
            (1001.0, 8 * 1024**3 + 1, False),
        ],
    )
    def test_within_limits(
        self, resource_manager_custom, estimated_time, estimated_space, expected
    ):
        """Test is_within_limits with various time and space estimates."""
        result = resource_manager_custom.is_within_limits(
            estimated_time, estimated_space
        )
        assert result == expected

    @pytest.mark.parametrize(
        ("estimated_time", "estimated_space"),
        [
            ("100", 8 * 1024**3),
            (500.0, "8e9"),
            ("500.0", "8e9"),
        ],
    )
    def test_within_limits_string_inputs(
        self, resource_manager_custom, estimated_time, estimated_space
    ):
        """Test is_within_limits with string representations of numbers."""
        result = resource_manager_custom.is_within_limits(
            estimated_time, estimated_space
        )
        expected = (
            float(estimated_time) <= resource_manager_custom.max_time
            and float(estimated_space) <= resource_manager_custom.max_space
        )
        assert result == expected

    @pytest.mark.parametrize(
        ("estimated_time", "estimated_space"),
        [
            (None, 8 * 1024**3),
            (500.0, None),
            (None, None),
            ([], {}),
            ({}, []),
        ],
    )
    def test_within_limits_invalid_types(
        self, resource_manager_custom, estimated_time, estimated_space
    ):
        """Test is_within_limits with invalid input types."""
        result = resource_manager_custom.is_within_limits(
            estimated_time, estimated_space
        )
        assert result is False

    @pytest.mark.parametrize(
        ("estimated_time", "estimated_space"),
        [
            (1e400, 8 * 1024**3),  # Overflow for time
            (500.0, 1e400),  # Overflow for space
            (1e400, 1e400),  # Overflow for both
        ],
    )
    def test_within_limits_overflow(
        self, resource_manager_custom, estimated_time, estimated_space
    ):
        """Test is_within_limits with overflow values."""
        result = resource_manager_custom.is_within_limits(
            estimated_time, estimated_space
        )
        assert result is False

    @pytest.mark.parametrize(
        ("estimated_time", "estimated_space"),
        [
            ("invalid", 8 * 1024**3),
            (500.0, "invalid"),
            ("invalid", "invalid"),
        ],
    )
    def test_within_limits_value_error(
        self, resource_manager_custom, estimated_time, estimated_space
    ):
        """Test is_within_limits with values that cannot be converted to float."""
        result = resource_manager_custom.is_within_limits(
            estimated_time, estimated_space
        )
        assert result is False


class TestResourceManagerUpdateAndGetLimits:
    def test_update_limits(self, resource_manager_default):
        """Test updating the resource limits."""
        new_time = 2000.0
        new_space = 12 * 1024**3  # 12 GB
        resource_manager_default.update_limits(new_time, new_space)
        limits = resource_manager_default.get_limits()
        assert limits["max_time"] == new_time
        assert limits["max_space"] == new_space

    def test_get_limits(self, resource_manager_custom):
        """Test retrieving the current resource limits."""
        limits = resource_manager_custom.get_limits()
        assert limits["max_time"] == 1000.0
        assert limits["max_space"] == 8 * 1024**3


# ==============================
# Tests for NetworkXGraphStore
# ==============================


class TestNetworkXGraphStoreGetNode:
    def test_get_existing_node(self, graph_store_simple):
        """Test retrieving an existing node."""
        node = graph_store_simple.get_node("A")
        assert node == {"id": "A", "data": "Alpha"}

    def test_get_non_existing_node(self, graph_store_simple):
        """Test retrieving a non-existing node."""
        node = graph_store_simple.get_node("C")
        assert node is None

    @pytest.mark.parametrize(
        "node_id",
        [
            123,  # Non-string ID
            None,  # None as ID
            ["A"],  # List as ID
            {"id": "A"},  # Dict as ID
        ],
    )
    def test_get_node_invalid_id(self, graph_store_simple, node_id):
        """Test retrieving a node with invalid node_id types."""
        node = graph_store_simple.get_node(node_id)
        assert node is None


class TestNetworkXGraphStoreGetNodes:
    def test_get_multiple_existing_nodes(self, graph_store_simple):
        """Test retrieving multiple existing nodes."""
        nodes = graph_store_simple.get_nodes(["A", "B"])
        expected = {
            "A": {"id": "A", "data": "Alpha"},
            "B": {"id": "B", "data": "Beta"},
        }
        assert nodes == expected

    def test_get_multiple_nodes_with_some_non_existing(self, graph_store_simple):
        """Test retrieving multiple nodes where some do not exist."""
        nodes = graph_store_simple.get_nodes(["A", "C"])
        expected = {
            "A": {"id": "A", "data": "Alpha"},
        }
        assert nodes == expected

    def test_get_nodes_empty_list(self, graph_store_simple):
        """Test retrieving nodes with an empty list."""
        nodes = graph_store_simple.get_nodes([])
        assert nodes == {}

    @pytest.mark.parametrize(
        "node_ids",
        [
            [123],  # Non-string ID
            [None],  # None as ID
            [["A"]],  # List within list
            [{"id": "A"}],  # Dict within list
            ["A", 123, None, {"id": "B"}],
        ],
    )
    def test_get_nodes_invalid_ids(self, graph_store_simple, node_ids):
        """Test retrieving nodes with invalid node_id types."""
        nodes = graph_store_simple.get_nodes(node_ids)
        if "A" in node_ids:
            assert nodes == {"A": {"id": "A", "data": "Alpha"}}
        else:
            assert nodes == {}


class TestNetworkXGraphStoreGetRel:
    def test_get_existing_relationship_simple_graph(self, graph_store_simple):
        """Test retrieving an existing relationship in a simple graph."""
        rel = graph_store_simple.get_rel("('A', 'B')")
        expected = {
            "id": "('A', 'B')",
            "source": "A",
            "target": "B",
            "relation": "connected",
        }
        assert rel == expected

    def test_get_non_existing_relationship_simple_graph(self, graph_store_simple):
        """Test retrieving a non-existing relationship in a simple graph."""
        rel = graph_store_simple.get_rel("('A', 'C')")
        assert rel is None

    def test_get_existing_relationship_multigraph(self, graph_store_multi):
        """Test retrieving an existing relationship in a multigraph."""
        rel = graph_store_multi.get_rel("('A', 'B')")
        expected = [
            {
                "id": "('A', 'B')-0",
                "source": "A",
                "target": "B",
                "relation": "connected",
            },
            {
                "id": "('A', 'B')-1",
                "source": "A",
                "target": "B",
                "relation": "linked",
            },
        ]
        assert rel == expected

    def test_get_non_existing_relationship_multigraph(self, graph_store_multi):
        """Test retrieving a non-existing relationship in a multigraph."""
        rel = graph_store_multi.get_rel("('A', 'C')")
        assert rel is None

    @pytest.mark.parametrize(
        "rel_id",
        [
            "invalid",  # Not a tuple
            "('A'",  # Incomplete tuple
            "('A', 'B', 'C')",  # Extra elements
            "",  # Empty string
            "123",  # Numeric string
            "None",
        ],
    )
    def test_get_rel_invalid_id(self, graph_store_simple, rel_id):
        """Test retrieving a relationship with invalid rel_id formats."""
        rel = graph_store_simple.get_rel(rel_id)
        assert rel is None


class TestNetworkXGraphStoreGetRels:
    def test_get_multiple_existing_rels_simple_graph(self, graph_store_simple):
        """Test retrieving multiple existing relationships in a simple graph."""
        rels = graph_store_simple.get_rels(["('A', 'B')"])
        expected = {
            "('A', 'B')": {
                "id": "('A', 'B')",
                "source": "A",
                "target": "B",
                "relation": "connected",
            }
        }
        assert rels == expected

    def test_get_multiple_existing_rels_multigraph(self, graph_store_multi):
        """Test retrieving multiple existing relationships in a multigraph."""
        rels = graph_store_multi.get_rels(["('A', 'B')"])
        expected = {
            "('A', 'B')": [
                {
                    "id": "('A', 'B')-0",
                    "source": "A",
                    "target": "B",
                    "relation": "connected",
                },
                {
                    "id": "('A', 'B')-1",
                    "source": "A",
                    "target": "B",
                    "relation": "linked",
                },
            ]
        }
        assert rels == expected

    def test_get_multiple_rels_with_some_non_existing(self, graph_store_simple):
        """Test retrieving multiple relationships where some do not exist."""
        rels = graph_store_simple.get_rels(["('A', 'B')", "('A', 'C')"])
        expected = {
            "('A', 'B')": {
                "id": "('A', 'B')",
                "source": "A",
                "target": "B",
                "relation": "connected",
            }
        }
        assert rels == expected

    def test_get_rels_empty_list(self, graph_store_simple):
        """Test retrieving relationships with an empty list."""
        rels = graph_store_simple.get_rels([])
        assert rels == {}

    @pytest.mark.parametrize(
        "rel_ids",
        [
            ["invalid"],  # Not a tuple
            ["('A'"],  # Incomplete tuple
            ["('A', 'B', 'C')"],  # Extra elements
            [""],  # Empty string
            ["123"],  # Numeric string
            ["None"],
            ["('A', 'B')", "invalid", "('A', 'C')"],
        ],
    )
    def test_get_rels_invalid_ids(self, graph_store_simple, rel_ids):
        """Test retrieving relationships with invalid rel_id formats."""
        rels = graph_store_simple.get_rels(rel_ids)
        expected = (
            {
                "('A', 'B')": {
                    "id": "('A', 'B')",
                    "source": "A",
                    "target": "B",
                    "relation": "connected",
                }
            }
            if "('A', 'B')" in rel_ids
            else {}
        )
        assert rels == expected


# ==============================
# Performance-Related Tests
# ==============================


@pytest.mark.timeout(2)  # Ensure that the test completes within 2 seconds
def test_resource_manager_performance(resource_manager_default):
    """Test that ResourceManager methods perform within acceptable time."""
    # This test ensures that methods execute quickly; actual performance testing would
    # require more setup
    assert resource_manager_default.get_limits() is not None
    assert resource_manager_default.is_within_limits(100, 1000) is not None


@pytest.mark.timeout(2)
def test_graph_store_performance(graph_store_simple):
    """Test that NetworkXGraphStore methods perform within acceptable time."""
    # Similar to above, ensure methods execute quickly
    assert graph_store_simple.get_node("A") is not None
    assert graph_store_simple.get_nodes(["A", "B"]) is not None
    assert graph_store_simple.get_rel("('A', 'B')") is not None
    assert graph_store_simple.get_rels(["('A', 'B')"]) is not None


# ==============================
# Boundary Condition Tests
# ==============================


class TestResourceManagerBoundaries:
    def test_zero_limits(self):
        """Test ResourceManager with zero limits."""
        rm = ResourceManager(max_time=0.0, max_space=0)
        assert rm.is_within_limits(0, 0) is True
        assert rm.is_within_limits(0.1, 0) is False
        assert rm.is_within_limits(0, 0.1) is False

    def test_negative_limits(self):
        """Test ResourceManager with negative limits."""
        with pytest.raises(ValueError, match="max_time must be non-negative"):
            ResourceManager(max_time=-100.0, max_space=-1)


class TestNetworkXGraphStoreBoundaries:
    def test_get_nodes_boundary(self, graph_store_simple):
        """Test get_nodes with boundary inputs."""
        # Test with a large list of node IDs including existing nodes "A" and "B"
        large_node_ids = [f"Node_{i}" for i in range(1000)] + ["A", "B"]
        nodes = graph_store_simple.get_nodes(large_node_ids)
        # Only "A" and "B" should be returned
        expected = {
            "A": {"id": "A", "data": "Alpha"},
            "B": {"id": "B", "data": "Beta"},
        }
        assert nodes == expected

    def test_get_rels_boundary(self, graph_store_simple):
        """Test get_rels with boundary inputs."""
        # Test with a large list of rel_ids
        large_rel_ids = ["('A', 'B')"] * 1000
        rels = graph_store_simple.get_rels(large_rel_ids)
        expected = {
            "('A', 'B')": {
                "id": "('A', 'B')",
                "source": "A",
                "target": "B",
                "relation": "connected",
            }
        }
        assert rels == expected


# ==============================
# Additional Edge Case Tests
# ==============================


class TestResourceManagerAdditionalEdges:
    def test_is_within_limits_exact_boundary(self, resource_manager_custom):
        """Test is_within_limits exactly on the boundary."""
        result = resource_manager_custom.is_within_limits(1000.0, 8 * 1024**3)
        assert result is True

    def test_is_within_limits_just_below_boundary(self, resource_manager_custom):
        """Test is_within_limits just below the boundary."""
        result = resource_manager_custom.is_within_limits(999.999, 8 * 1024**3 - 1)
        assert result is True

    def test_is_within_limits_just_above_boundary(self, resource_manager_custom):
        """Test is_within_limits just above the boundary."""
        result = resource_manager_custom.is_within_limits(1000.001, 8 * 1024**3 + 1)
        assert result is False


class TestNetworkXGraphStoreAdditionalEdges:
    def test_get_rel_with_extra_whitespace(self, graph_store_simple):
        """Test get_rel with rel_id containing extra whitespace."""
        rel = graph_store_simple.get_rel(" ( 'A' , 'B' ) ")
        expected = {
            "id": " ( 'A' , 'B' ) ",
            "source": "A",
            "target": "B",
            "relation": "connected",
        }
        assert rel == expected

    def test_get_rels_duplicate_ids(self, graph_store_multi):
        """Test get_rels with duplicate relationship IDs in a multigraph."""
        rels = graph_store_multi.get_rels(["('A', 'B')", "('A', 'B')"])
        expected = {
            "('A', 'B')": [
                {
                    "id": "('A', 'B')-0",
                    "source": "A",
                    "target": "B",
                    "relation": "connected",
                },
                {
                    "id": "('A', 'B')-1",
                    "source": "A",
                    "target": "B",
                    "relation": "linked",
                },
            ]
        }
        assert rels == expected



import os

from transformers import pipeline

from nxlu.config import Intent
from nxlu.constants import GENERAL_EXPLORATION_PROMPT
from nxlu.utils.misc import most_common_element

if not os.environ.get("TOKENIZERS_PARALLELISM"):
    os.environ["TOKENIZERS_PARALLELISM"] = "true"

__all__ = [
    "IntentClassifier",
    "intent2prompt",
    "classify_domain",
]


class IntentClassifier:
    """Classify high-level intents from a user's query using zero-shot
    classification.

    Attributes
    ----------
    classifier : transformers.pipeline
        The zero-shot classification pipeline using the BART model.

    Methods
    -------
    fit_transform(query: str, threshold: float) -> list[Intent]
        Classifies the query and returns the list of predicted intents.
    """

    def __init__(self):
        """
        Initialize the IntentClassifier class.

        The classifier uses a zero-shot classification pipeline with the
        'facebook/bart-large-mnli' model.
        """
        self.classifier = pipeline(
            "zero-shot-classification", model="facebook/bart-large-mnli", device=-1
        )

    def fit_transform(self, query: str, threshold: float = 0.5) -> list[Intent]:
        """
        Classify the given query into high-level intents using zero-shot
        classification.

        Parameters
        ----------
        query : str
            The user query to classify.
        threshold : float, optional
            The confidence threshold for selecting intents (default is 0.5).

        Returns
        -------
        list[Intent]
            A list of `Intent` objects that have confidence scores higher
            than the threshold.
        """
        candidate_labels = [intent.value for intent in Intent]
        results = self.classifier(query, candidate_labels, multi_label=True)

        # select intents with softmax confidence above a threshold
        selected_intents = [
            Intent(label.strip())
            for label, score in zip(results["labels"], results["scores"])
            if score > threshold
        ]

        return selected_intents


def intent2prompt(intents: list[Intent]) -> str:
    """
    Map a list of intents to their corresponding analytical prompt.

    Parameters
    ----------
    intents : list of Intent
        The list of classified intents.

    Returns
    -------
    str
        A string representing the appropriate analytical prompt based on
        the first matching intent. Defaults to 'GENERAL_EXPLORATION' if no
        match is found.
    """
    intent_to_analysis = {
        Intent.EXPLORATION: "GENERAL_EXPLORATION",
        Intent.CAUSAL_EXPLANATION: "PATH_ANALYSIS",
        Intent.PATTERN_RECOGNITION: "COMMUNITY_DETECTION",
        Intent.TREND_ANALYSIS: "TEMPORAL_ANALYSIS",
        Intent.CLASSIFICATION: "CENTRALITY_ANALYSIS",
        Intent.ERROR_DETECTION: "ANOMALY_DETECTION",
    }

    for intent in intents:
        if intent in intent_to_analysis:
            return intent_to_analysis[intent]

    return "GENERAL_EXPLORATION"


def classify_domain(attributes: dict) -> str:
    """Classify the domain based on node and edge attributes.

    Parameters
    ----------
    attributes : dict
        A dictionary containing attributes for nodes and edges. The
        'query' field is optional.

    Returns
    -------
    str
        The most common predicted domain based on the node and edge
        attributes and optional query.
    """
    import torch
    from transformers import AutoConfig, AutoTokenizer

    from nxlu.processing.embed import CustomModel

    config = AutoConfig.from_pretrained("nvidia/domain-classifier")
    tokenizer = AutoTokenizer.from_pretrained("nvidia/domain-classifier")
    model = CustomModel.from_pretrained("nvidia/domain-classifier")

    node_attributes = [str(node[1]) for node in attributes["nodes"]]
    edge_attributes = [str(edge) for edge in attributes["edges"]]

    input_strings = node_attributes + edge_attributes

    if attributes.get("query") != GENERAL_EXPLORATION_PROMPT:
        input_strings.append(attributes["query"])

    inputs = tokenizer(
        input_strings,
        return_tensors="pt",
        padding="longest",
        truncation=True,
    )
    outputs = model(inputs["input_ids"], inputs["attention_mask"])

    predicted_classes = torch.argmax(outputs, dim=1).cpu().numpy()
    predicted_domains = [config.id2label[class_idx] for class_idx in predicted_classes]

    return most_common_element(predicted_domains)



import random

import numpy as np
import pytest
import torch

from nxlu.utils.misc import (
    ReadableDict,
    cosine_similarity,
    most_common_element,
    normalize_name,
    parse_algorithms,
    sanitize_input,
    scrub_braces,
    set_seed,
)


@pytest.fixture
def sample_context():
    return {
        "name": "John Doe",
        "description": "{A sample description}",
        "nested": {"message": "{A message with {nested} braces}"},
        "tags": ["{tag1}", "{tag2}", "{tag3}"],
        "numbers": (1, 2, "{3}", 4),
        "empty_dict": {},
    }


@pytest.mark.parametrize(
    ("query", "expected_output"),
    [
        ("<script>alert('XSS')</script>", "scriptalert('XSS')script"),
        ("<<>>", ""),
        ("<tag>", "tag"),
        ("normal query", "normal query"),
        ("", ""),
    ],
)
def test_sanitize_input(query, expected_output):
    assert sanitize_input(query) == expected_output


# Test for most_common_element
@pytest.mark.parametrize(
    ("elements", "expected_output"),
    [
        ([1, 2, 2, 3], 2),  # Normal case with most common element
        (["apple", "banana", "apple"], "apple"),  # Most common string
        ([True, False, True], True),  # Booleans
        ([None, None, None], None),  # All None
        ([], None),  # Empty list
        ([1, 1, 2, 2], 1),  # Equal frequency, returns the first
        (["single"], "single"),  # Single element list
    ],
)
def test_most_common_element(elements, expected_output):
    assert most_common_element(elements) == expected_output


# Test for edge cases in most_common_element
def test_most_common_element_invalid_input():
    with pytest.raises(TypeError):
        most_common_element(1234)  # Not a list input


# Test for scrub_braces
@pytest.mark.parametrize(
    ("context", "expected_output"),
    [
        # Basic test cases
        ({"key": "{value}"}, {"key": "value"}),
        ({"key": "no_braces"}, {"key": "no_braces"}),  # No braces to scrub
        ({"key": "{{double braces}}"}, {"key": "{{double braces}}"}),  # Double braces
        ({"key": "{}"}, {"key": ""}),  # Empty braces
        # Nested dictionary
        ({"key": {"subkey": "{subvalue}"}}, {"key": {"subkey": "subvalue"}}),
        # List inside dictionary
        ({"key": ["{listitem1}", "{listitem2}"]}, {"key": ["listitem1", "listitem2"]}),
        # Tuple inside dictionary
        (
            {"key": ("{tupleitem1}", "{tupleitem2}")},
            {"key": ("tupleitem1", "tupleitem2")},
        ),
        # Set inside dictionary
        ({"key": {"{setitem1}", "{setitem2}"}}, {"key": {"setitem1", "setitem2"}}),
    ],
)
def test_scrub_braces(context, expected_output):
    assert scrub_braces(context) == expected_output


def test_scrub_braces_complex_structure(sample_context):
    expected_output = {
        "name": "John Doe",
        "description": "A sample description",
        "nested": {"message": "A message with nested braces"},
        "tags": ["tag1", "tag2", "tag3"],
        "numbers": (1, 2, "3", 4),
        "empty_dict": {},
    }
    assert scrub_braces(sample_context) == expected_output


def test_scrub_braces_empty():
    assert scrub_braces({}) == {}


@pytest.mark.parametrize(
    "invalid_context",
    [
        None,
        1234,
        "string",
        [1, 2, 3],
    ],
)
def test_scrub_braces_invalid_input(invalid_context):
    assert scrub_braces(invalid_context) == {}


def test_readable_dict_and_normalize_name():
    data = ReadableDict(
        {
            "simple_key": "simple_value",
            "nested_dict": {
                "inner_key1": "value1",
                "inner_key2": "value2",
            },
            "long_string": (
                "This is a very long string that should be wrapped properly by the "
                "ReadableDict's __str__ method.\n"
                "This is a very long string that should be wrapped properly by the "
                "ReadableDict's __str__ method.\n"
            ),
        }
    )

    result_str = str(data)

    assert "{\n" in result_str
    assert "'simple_key': \"simple_value\"" in result_str
    assert "'nested_dict': {\n" in result_str
    assert "'inner_key1': \"value1\"" in result_str
    assert "'inner_key2': \"value2\"" in result_str
    assert "'long_string': '''" in result_str
    assert "This is a very long string" in result_str
    assert "'''" in result_str

    test_cases = [
        ("Best-Partition_Index", "louvain"),
        ("Girvan-Newman", "girvan newman"),
        ("K-Means", "k means"),
        ("Coefficient-Index", ""),
        ("unknown_algorithm", "unknown algorithm"),
        ("KMeans", "k means"),
    ]
    for name, expected in test_cases:
        assert normalize_name(name) == expected


def test_cosine_similarity_set_seed_parse_algorithms():
    X = np.array([[1, 0], [0, 1]])
    Y = np.array([[1, 0], [1, 1]])
    expected_similarity = np.array(
        [
            [1.0, 1 / np.sqrt(2)],
            [0.0, 1 / np.sqrt(2)],
        ]
    )
    computed_similarity = cosine_similarity(X, Y)
    np.testing.assert_almost_equal(computed_similarity, expected_similarity)

    set_seed(123)

    random_num = random.randint(0, 100)
    assert random_num == 6

    torch_random_num = torch.rand(1).item()
    assert np.isclose(torch_random_num, 0.29611194133758545, atol=1e-7)


def test_parse_algorithms():
    sample_encyclopedia = {
        "Louvain": {"algorithm_category": "Community Detection"},
        "Girvan-Newman": {"algorithm_category": "Community Detection"},
        "KMeans": {"algorithm_category": "Clustering"},
    }
    supported, categories, standardized = parse_algorithms(
        sample_encyclopedia, normalize_name
    )
    assert supported == ["Louvain", "Girvan-Newman", "KMeans"]
    assert categories == {
        "louvain": "Community Detection",
        "girvan newman": "Community Detection",
        "k means": "Clustering",
    }
    assert standardized == ["louvain", "girvan newman", "k means"]



import logging
import sys
from logging.handlers import TimedRotatingFileHandler
from typing import Any

from _nxlu.config import LoggerConfig, LoggingConfig, LoggingHandlerConfig, _config

__all__ = [
    "setup_logger",
    "create_handler",
    "setup_logger_from_config",
    "update_logger",
    "on_config_change",
    "initialize_logging",
    "get_default_logger",
    "disable_logger",
]


def setup_logger(logger_cfg: LoggerConfig) -> None:
    """Set up a single logger based on LoggerConfig."""
    logger = logging.getLogger(logger_cfg.name)
    logger.setLevel(getattr(logging, logger_cfg.level.upper(), logging.INFO))

    logger.handlers = []

    for handler_cfg in logger_cfg.handlers:
        handler = create_handler(handler_cfg)
        if handler:
            logger.addHandler(handler)


def create_handler(handler_cfg: LoggingHandlerConfig) -> logging.Handler | None:
    """Create a logging handler based on the handler configuration."""
    formatter = logging.Formatter(handler_cfg.formatter)
    handler = None

    if handler_cfg.handler_type.lower() == "console":
        handler = logging.StreamHandler(sys.stdout)
    elif handler_cfg.handler_type.lower() == "file":
        if not handler_cfg.log_file:
            raise ValueError("log_file must be specified for file handlers.")
        if handler_cfg.rotate_logs:
            handler = TimedRotatingFileHandler(
                handler_cfg.log_file,
                when=handler_cfg.when,
                backupCount=handler_cfg.backup_count,
            )
        else:
            handler = logging.FileHandler(handler_cfg.log_file)
    else:
        raise ValueError(f"Unsupported handler type: {handler_cfg.handler_type}")

    if handler:
        handler.setLevel(getattr(logging, handler_cfg.level.upper(), logging.INFO))
        handler.setFormatter(formatter)

    return handler


def setup_logger_from_config(logging_config: LoggingConfig) -> None:
    """Set up all loggers based on the provided LoggingConfig."""
    for logger_cfg in logging_config.loggers:
        setup_logger(logger_cfg)


def update_logger(name: str, action: str, value: Any = None) -> None:
    """Update a specific logger based on action.

    Parameters
    ----------
    name : str
        The name of the logger to update.
    action : str
        The action to perform: 'add_logger', 'update_logger', 'remove_logger',
        'verbosity_level'.
    value : Any, optional
        Additional value needed for certain actions (e.g., verbosity level).
    """
    if action in ("add_logger", "update_logger"):
        logger_cfg = next(
            (
                logger
                for logger in _config.logging_config.loggers
                if logger.name == name
            ),
            None,
        )
        if logger_cfg:
            setup_logger(logger_cfg)
    elif action == "remove_logger":
        logger = logging.getLogger(name)
        logger.handlers = []
    elif action == "verbosity_level":
        verbosity_level = value  # value is expected to be an integer (0, 1, 2)
        logger_name = "nxlu"
        logger = logging.getLogger(logger_name)

        if verbosity_level == 0:
            logger.handlers = []
            logger.disabled = True
        else:
            logger.disabled = False
            level_map = {1: "INFO", 2: "DEBUG"}
            log_level = level_map.get(verbosity_level, "INFO")
            logger.setLevel(getattr(logging, log_level.upper(), logging.INFO))

            if not logger.handlers:
                handler = logging.StreamHandler(sys.stdout)
                handler.setLevel(getattr(logging, log_level.upper(), logging.INFO))
                formatter = logging.Formatter(
                    "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
                )
                handler.setFormatter(formatter)
                logger.addHandler(handler)
            else:
                for handler in logger.handlers:
                    handler.setLevel(getattr(logging, log_level.upper(), logging.INFO))
    else:
        raise ValueError(f"Unknown action: {action}")


def on_config_change(name: str, value: Any) -> None:
    """Handle configuration changes.

    Parameters
    ----------
    name : str
        The name of the configuration parameter that changed.
    value : any
        The new value of the configuration parameter.
    """
    if name == "verbosity_level":
        update_logger("nxlu", "verbosity_level", value)
    elif name in ("add_logger", "update_logger", "remove_logger"):
        update_logger(name, name, value)
    else:
        pass


def initialize_logging() -> None:
    """Initialize logging based on the current configuration."""
    setup_logger_from_config(_config.logging_config)
    _config.register_observer(on_config_change)


def get_default_logger() -> logging.Logger:
    """Return a default logger instance for simple logging."""
    return logging.getLogger("nxlu")


def disable_logger(logger_name: str) -> None:
    """Disables the logger by removing all handlers."""
    logger = logging.getLogger(logger_name)
    logger.disabled = True
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)



from _nxlu.config import NxluConfig
from nxlu.core.base import BaseBackend
from nxlu.core.factory import get_backend

__all__ = ["BackendManager"]


class BackendManager:
    """Manage the NetworkX backend for nxlu.

    Parameters
    ----------
    config : NxluConfig
        Configuration settings for the model and backend.
    """

    def __init__(self, config: NxluConfig):
        """Initialize the BackendManager with the given configuration.

        Parameters
        ----------
        config : NxluConfig
            Configuration settings for the model and backend.
        """
        self.config = config
        self.backend = get_backend(config.get_backend_name())

    def get_backend(self) -> BaseBackend:
        """Retrieve the instantiated backend.

        Returns
        -------
        BaseBackend
            The instantiated backend.
        """
        return self.backend



from _nxlu.config import _config

__all__ = ["get_info"]


def get_info():
    """Provide information about the backend.

    Returns
    -------
    Dict[str, Any]
        A dictionary containing metadata about the backend.
    """
    return {
        "name": "Network Language Understanding",
        "version": "1.0",
        "description": (
            "A backend that provides natural language explanations of graph network "
            "topology"
        ),
        "author": "Derek Alexander",
        "url": "https://github.com/dpys/gar",
        "graph_class": "gar.interface:LLMGraph",
        "functions": {
            "approximate_current_flow_betweenness_centrality": {},
            "betweenness_centrality": {},
            "betweenness_centrality_subset": {},
            "closeness_centrality": {},
            "communicability_betweenness_centrality": {},
            "current_flow_betweenness_centrality": {},
            "current_flow_betweenness_centrality_subset": {},
            "current_flow_closeness_centrality": {},
            "degree_centrality": {},
            "dispersion": {},
            "edge_betweenness_centrality": {},
            "edge_betweenness_centrality_subset": {},
            "edge_current_flow_betweenness_centrality": {},
            "edge_current_flow_betweenness_centrality_subset": {},
            "edge_load_centrality": {},
            "eigenvector_centrality": {},
            "eigenvector_centrality_numpy": {},
            "estrada_index": {},
            "global_reaching_centrality": {},
            "group_betweenness_centrality": {},
            "group_closeness_centrality": {},
            "group_degree_centrality": {},
            "group_in_degree_centrality": {},
            "group_out_degree_centrality": {},
            "harmonic_centrality": {},
            "in_degree_centrality": {},
            "incremental_closeness_centrality": {},
            "information_centrality": {},
            "katz_centrality": {},
            "katz_centrality_numpy": {},
            "laplacian_centrality": {},
            "load_centrality": {},
            "local_reaching_centrality": {},
            "out_degree_centrality": {},
            "percolation_centrality": {},
            "prominent_group": {},
            "second_order_centrality": {},
            "subgraph_centrality": {},
            "subgraph_centrality_exp": {},
            "trophic_differences": {},
            "trophic_incoherence_parameter": {},
            "trophic_levels": {},
            "voterank": {},
            "enumerate_all_cliques": {},
            "max_weight_clique": {},
            "node_clique_number": {},
            "number_of_cliques": {},
            "average_clustering": {},
            "clustering": {},
            "generalized_degree": {},
            "square_clustering": {},
            "transitivity": {},
            "triangles": {},
            "articulation_points": {},
            "attracting_components": {},
            "biconnected_component_edges": {},
            "biconnected_components": {},
            "condensation": {},
            "density": {},
            "connected_components": {},
            "kosaraju_strongly_connected_components": {},
            "node_connected_component": {},
            "number_attracting_components": {},
            "number_connected_components": {},
            "number_strongly_connected_components": {},
            "number_weakly_connected_components": {},
            "strongly_connected_components": {},
            "strongly_connected_components_recursive": {},
            "weakly_connected_components": {},
            "all_node_cuts": {},
            "all_pairs_node_connectivity": {},
            "average_node_connectivity": {},
            "bridge_components": {},
            "edge_connectivity": {},
            "edge_disjoint_paths": {},
            "k_components": {},
            "k_edge_augmentation": {},
            "k_edge_components": {},
            "k_edge_subgraphs": {},
            "local_edge_connectivity": {},
            "local_node_connectivity": {},
            "minimum_edge_cut": {},
            "minimum_node_cut": {},
            "minimum_st_edge_cut": {},
            "minimum_st_node_cut": {},
            "node_connectivity": {},
            "node_disjoint_paths": {},
            "stoer_wagner": {},
            "chordless_cycles": {},
            "cycle_basis": {},
            "girth": {},
            "minimum_cycle_basis": {},
            "recursive_simple_cycles": {},
            "simple_cycles": {},
            "boykov_kolmogorov": {},
            "capacity_scaling": {},
            "cost_of_flow": {},
            "dinitz": {},
            "edmonds_karp": {},
            "gomory_hu_tree": {},
            "max_flow_min_cost": {},
            "maximum_flow": {},
            "maximum_flow_value": {},
            "min_cost_flow": {},
            "min_cost_flow_cost": {},
            "minimum_cut": {},
            "minimum_cut_value": {},
            "network_simplex": {},
            "preflow_push": {},
            "shortest_augmenting_path": {},
            "all_pairs_all_shortest_paths": {},
            "all_pairs_bellman_ford_path": {},
            "all_pairs_bellman_ford_path_length": {},
            "all_pairs_dijkstra": {},
            "all_pairs_dijkstra_path": {},
            "all_pairs_dijkstra_path_length": {},
            "all_pairs_shortest_path": {},
            "all_pairs_shortest_path_length": {},
            "all_shortest_paths": {},
            "astar_path": {},
            "astar_path_length": {},
            "average_shortest_path_length": {},
            "bellman_ford_path": {},
            "bellman_ford_path_length": {},
            "bellman_ford_predecessor_and_distance": {},
            "bidirectional_dijkstra": {},
            "bidirectional_shortest_path": {},
            "dijkstra_path": {},
            "dijkstra_path_length": {},
            "dijkstra_predecessor_and_distance": {},
            "floyd_warshall": {},
            "floyd_warshall_numpy": {},
            "floyd_warshall_predecessor_and_distance": {},
            "goldberg_radzik": {},
            "johnson": {},
            "multi_source_dijkstra": {},
            "multi_source_dijkstra_path": {},
            "multi_source_dijkstra_path_length": {},
            "negative_edge_cycle": {},
            "predecessor": {},
            "reconstruct_path": {},
            "shortest_path": {},
            "shortest_path_length": {},
            "single_source_all_shortest_paths": {},
            "single_source_bellman_ford": {},
            "single_source_bellman_ford_path": {},
            "single_source_bellman_ford_path_length": {},
            "single_source_dijkstra": {},
            "single_source_dijkstra_path": {},
            "single_source_dijkstra_path_length": {},
            "single_source_shortest_path": {},
            "single_source_shortest_path_length": {},
            "single_target_shortest_path": {},
            "single_target_shortest_path_length": {},
            "asyn_fluidc": {},
            "asyn_lpa_communities": {},
            "edge_betweenness_partition": {},
            "edge_current_flow_betweenness_partition": {},
            "fast_label_propagation_communities": {},
            "girvan_newman": {},
            "greedy_modularity_communities": {},
            "k_clique_communities": {},
            "kernighan_lin_bisection": {},
            "label_propagation_communities": {},
            "louvain_communities": {},
            "louvain_partitions": {},
            "lukes_partitioning": {},
            "modularity": {},
            "naive_greedy_modularity_communities": {},
            "partition_quality": {},
            "equitable_color": {},
            "greedy_color": {},
            "strategy_connected_sequential": {},
            "strategy_connected_sequential_bfs": {},
            "strategy_connected_sequential_dfs": {},
            "strategy_independent_set": {},
            "strategy_largest_first": {},
            "strategy_random_sequential": {},
            "strategy_saturation_largest_first": {},
            "strategy_smallest_last": {},
            "categorical_edge_match": {},
            "categorical_multiedge_match": {},
            "categorical_node_match": {},
            "could_be_isomorphic": {},
            "fast_could_be_isomorphic": {},
            "faster_could_be_isomorphic": {},
            "generic_edge_match": {},
            "generic_multiedge_match": {},
            "generic_node_match": {},
            "numerical_edge_match": {},
            "numerical_multiedge_match": {},
            "numerical_node_match": {},
            "rooted_tree_isomorphism": {},
            "tree_isomorphism": {},
            "vf2pp_all_isomorphisms": {},
            "vf2pp_is_isomorphic": {},
            "vf2pp_isomorphism": {},
            "is_connected": {},
            "matching_dict_to_set": {},
            "max_weight_matching": {},
            "maximal_matching": {},
            "min_weight_matching": {},
            "all_topological_sorts": {},
            "ancestors": {},
            "antichains": {},
            "compute_v_structures": {},
            "dag_longest_path": {},
            "dag_longest_path_length": {},
            "dag_to_branching": {},
            "descendants": {},
            "lexicographical_topological_sort": {},
            "root_to_leaf_paths": {},
            "topological_generations": {},
            "topological_sort": {},
            "transitive_closure": {},
            "transitive_closure_dag": {},
            "transitive_reduction": {},
            "branching_weight": {},
            "greedy_branching": {},
            "join": {},
            "join_trees": {},
            "junction_tree": {},
            "maximum_spanning_edges": {},
            "maximum_spanning_tree": {},
            "minimum_spanning_edges": {},
            "minimum_spanning_tree": {},
            "number_of_spanning_trees": {},
            "partition_spanning_tree": {},
            "random_spanning_tree": {},
            "asadpour_atsp": {},
            "christofides": {},
            "clique_removal": {},
            "diameter": {},
            "greedy_tsp": {},
            "large_clique_size": {},
            "max_clique": {},
            "maximum_independent_set": {},
            "metric_closure": {},
            "min_edge_dominating_set": {},
            "min_maximal_matching": {},
            "min_weighted_dominating_set": {},
            "min_weighted_vertex_cover": {},
            "one_exchange": {},
            "ramsey_R2": {},
            "randomized_partitioning": {},
            "simulated_annealing_tsp": {},
            "steiner_tree": {},
            "threshold_accepting_tsp": {},
            "traveling_salesman_problem": {},
            "treewidth_min_degree": {},
            "treewidth_min_fill_in": {},
            "google_matrix": {},
            "hits": {},
            "pagerank": {},
            "bfs_beam_edges": {},
            "bfs_edges": {},
            "bfs_labeled_edges": {},
            "bfs_layers": {},
            "bfs_predecessors": {},
            "bfs_successors": {},
            "bfs_tree": {},
            "descendants_at_distance": {},
            "dfs_edges": {},
            "dfs_labeled_edges": {},
            "dfs_postorder_nodes": {},
            "dfs_predecessors": {},
            "dfs_preorder_nodes": {},
            "dfs_successors": {},
            "dfs_tree": {},
            "edge_bfs": {},
            "edge_dfs": {},
            "generic_bfs_edges": {},
            "best_partition": {},
        },
        "default_config": _config,
    }



import logging
import warnings
from typing import Any

import networkx as nx

warnings.filterwarnings("ignore")


logger = logging.getLogger("nxlu")


__all__ = [
    "create_graph_from_data",
    "inspect_data",
    "filter_graph_by_relation",
    "construct_graph",
]


def create_graph_from_data(data: dict[str, Any]) -> nx.Graph:
    """Create a NetworkX graph from general-purpose data.

    Parameters
    ----------
    data : Dict[str, Any]
        Dictionary containing 'nodes' and 'edges' with their attributes.

    Returns
    -------
    nx.Graph
        The constructed NetworkX graph.
    """
    graph = nx.Graph()

    for node in data.get("nodes", []):
        node_id = node.get("id")
        if node_id is None:
            logger.warning("A node without an 'id' was found and skipped.")
            continue
        attributes = node.get("attributes", {})
        graph.add_node(node_id, **attributes)

    for edge in data.get("edges", []):
        source = edge.get("source")
        target = edge.get("target")
        if source is None or target is None:
            logger.warning(f"Incomplete edge data {edge} found and skipped.")
            continue
        attributes = edge.get("attributes", {})
        graph.add_edge(source, target, **attributes)

    return graph


def validate_graph(G: nx.Graph) -> bool:
    for u, v in G.edges():
        if not G.has_node(u) or not G.has_node(v):
            logger.error(f"Invalid edge detected between {u} and {v}.")
            return False
    logger.info("Graph validation successful.")
    return True


def inspect_data(dataset: dict[str, Any]) -> tuple[Any, Any]:
    """Inspect and print the types and samples of 'nodes' and 'links' in the dataset.

    Parameters
    ----------
    dataset : Dict[str, Any]
        The dataset containing 'nodes' and 'links'.

    Returns
    -------
    Tuple[Any, Any]
        The 'nodes' and 'links' extracted from the dataset.
    """
    nodes = dataset.get("nodes")
    links = dataset.get("links")

    if isinstance(nodes, list):
        for _node in nodes[:5]:
            pass
    elif isinstance(nodes, dict):
        for _key, _value in list(nodes.items())[:5]:
            pass
    else:
        pass

    if isinstance(links, list):
        for _link in links[:5]:
            pass
    elif isinstance(links, dict):
        for _key, _value in list(links.items())[:5]:
            pass
    else:
        pass

    return nodes, links


def filter_graph_by_relation(
    graph: nx.MultiDiGraph, desired_relation: str = "phenotype_phenotype"
) -> nx.Graph:
    """Filter the graph to include only edges with the specified 'relation' attribute.

    Parameters
    ----------
    graph : nx.MultiDiGraph
        The original graph.
    desired_relation : str, optional
        The relation attribute to filter edges by, by default "phenotype_phenotype".

    Returns
    -------
    nx.Graph
        The filtered graph containing only the desired edges.
    """
    filtered_graph = nx.Graph()
    filtered_graph.graph.update(graph.graph)

    for u, v, key, data in graph.edges(keys=True, data=True):
        if data.get("relation") == desired_relation:
            filtered_graph.add_edge(u, v, key=key, **data)

    return filtered_graph


def construct_graph(nodes: Any, links: Any) -> nx.Graph:
    """Construct a NetworkX graph from nodes and links.

    Parameters
    ----------
    nodes : Any
        List or dict of nodes with optional attributes.
    links : Any
        List or dict of links with optional attributes.

    Returns
    -------
    nx.Graph
        The constructed NetworkX graph.
    """
    graph = nx.Graph()

    if isinstance(nodes, list):
        if not nodes:
            logger.warning("'nodes' list is empty.")
        elif isinstance(nodes[0], dict):
            for node in nodes:
                node_id = node.get("id")
                weight = node.get("weight", 1)
                if node_id:
                    graph.add_node(node_id, weight=weight)
                else:
                    logger.warning("A node without an 'id' was found and skipped.")
        else:
            graph.add_nodes_from((node, {"weight": 1}) for node in nodes)
    elif isinstance(nodes, dict):
        graph.add_nodes_from(
            (node_id, {"weight": weight}) for node_id, weight in nodes.items()
        )
    else:
        logger.warning("Unrecognized 'nodes' structure. Nodes may not have weights.")
        try:
            graph.add_nodes_from((node, {"weight": 1}) for node in nodes)
        except Exception:
            logger.exception("Error adding nodes")

    if isinstance(links, list):
        if not links:
            logger.warning("'links' list is empty.")
        else:
            first_link = links[0]
            if isinstance(first_link, tuple):
                for link in links:
                    if isinstance(link, tuple):
                        if len(link) == 3:
                            source, target, weight = link
                            graph.add_edge(source, target, weight=weight)
                        elif len(link) == 2:
                            source, target = link
                            graph.add_edge(source, target, weight=1)
                        else:
                            logger.warning(
                                f"Unexpected link tuple size {len(link)}: {link}"
                            )
                    else:
                        logger.warning(f"Link is not a tuple: {link}")
            elif isinstance(first_link, dict):
                for link in links:
                    if isinstance(link, dict):
                        source = link.get("source")
                        target = link.get("target")
                        weight = link.get("weight", 1)
                        if source and target:
                            graph.add_edge(source, target, weight=weight)
                        else:
                            logger.warning(f"Incomplete link data {link}")
                    else:
                        logger.warning(f"Link is not a dict: {link}")
            else:
                logger.warning(f"Unexpected link type: {type(first_link)}")
    elif isinstance(links, dict):
        first_key = next(iter(links), None)
        if isinstance(first_key, tuple):
            for edge, attrs in links.items():
                if isinstance(edge, tuple) and len(edge) >= 2:
                    source, target = edge[:2]
                    weight = attrs.get("weight", 1) if isinstance(attrs, dict) else 1
                    graph.add_edge(source, target, weight=weight)
                else:
                    logger.warning(
                        f"Invalid edge format {edge} with attributes {attrs}"
                    )
        else:
            for source, targets in links.items():
                for target in targets:
                    graph.add_edge(source, target, weight=1)
    else:
        logger.warning("Unrecognized 'links' structure. No edges added.")

    return graph



import logging
import warnings

warnings.filterwarnings("ignore")

logger = logging.getLogger("nxlu")

__all__ = ["handle_error"]


def handle_error(e: Exception, message: str = "An error occurred") -> None:
    """Log an error message along with the exception details.

    Parameters
    ----------
    e : Exception
        The exception that was raised.
    message : str, optional
        A custom error message, by default "An error occurred".

    Returns
    -------
    None
    """
    logger.error(f"{message}: {e}")


def raise_backend_key_error(backend_name: str) -> None:
    """Raise a key error if the backend name is not found in the pyproject.toml"""
    raise KeyError(f"Backend '{backend_name}' not found in entry points.")



import asyncio
import inspect
import logging
import warnings
from typing import Any

import networkx as nx
from langchain.chains import LLMChain
from langchain.prompts import ChatPromptTemplate
from tenacity import retry, stop_after_attempt, wait_exponential

from nxlu.config import Framework, Intent, NxluConfig
from nxlu.constants import SYSTEM_PROMPT, AnalysisPrompts
from nxlu.explanation.classify import IntentClassifier, intent2prompt
from nxlu.explanation.interrogate import GraphInterrogator
from nxlu.io import load_algorithm_docs
from nxlu.processing.analyze import get_algorithm_function
from nxlu.utils.control import init_llm_model
from nxlu.utils.misc import sanitize_input

warnings.filterwarnings("ignore")


logger = logging.getLogger("nxlu")


__all__ = ["UnifiedLLMChain", "GraphExplainer"]

MAX_LLM_THREADS = 5  # limiting to 5 concurrent calls for now


class UnifiedLLMChain:
    """A unified wrapper to handle both LangChain and LlamaIndex LLMs."""

    def __init__(self, config: NxluConfig):
        """Initialize the UnifiedLLMChain with the appropriate LLM and framework."""
        self.config = config
        framework = config.get_llm_framework()
        self.llm = init_llm_model(config)

        self.system_prompt = SYSTEM_PROMPT

        if framework == Framework.LANGCHAIN:
            self.chat_prompt = ChatPromptTemplate.from_messages(
                [("system", self.system_prompt), ("user", "{prompt}")]
            )
            self.chain = LLMChain(llm=self.llm, prompt=self.chat_prompt)

    def llamaindex_complete(self, prompt: str) -> str:
        """Structure the system and user prompts and call the LlamaIndex LLM."""
        full_prompt = f"{self.system_prompt}\n\nUser Prompt: {prompt}"
        return self.llm.complete(full_prompt)

    async def _call_llm_chain(self, prompt: str) -> str | None:
        """Call the LLM chain depending on the framework."""
        framework = self.config.get_llm_framework()
        if framework == Framework.LANGCHAIN:
            return await self.chain.arun(prompt=prompt)
        if framework == Framework.LLAMAINDEX:
            return self.llamaindex_complete(prompt)

    def run(self, prompt: str) -> str | None:
        """Run the chain synchronously depending on the framework."""
        framework = self.config.get_llm_framework()
        if framework == Framework.LANGCHAIN:
            return self.chain.run(prompt=prompt)
        if framework == Framework.LLAMAINDEX:
            return self.llamaindex_complete(prompt)


class GraphExplainer:
    """A class to generate explanations for graphs based on user queries using LLM and
    graph interrogation.

    Parameters
    ----------
    config : NxluConfig
        The configuration object used to initialize the LLM and GraphInterrogator.
    """

    def __init__(self, config: NxluConfig):
        """Initialize the GraphExplainer with LLM and GraphInterrogator."""
        self.config = config
        self.llm = init_llm_model(self.config)
        self.intent_classifier = None
        self.interrogator = None
        self._llm_semaphore = asyncio.Semaphore(MAX_LLM_THREADS)

        self.chain = UnifiedLLMChain(self.config)

    @retry(
        stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=4, max=10)
    )
    async def _call_llm_chain(self, prompt: str) -> str | None:
        async with self._llm_semaphore:
            return await self.chain._call_llm_chain(prompt)

    async def explain_async(
        self,
        graph_data: nx.Graph,
        query: str | None = None,
    ) -> str:
        """Generate an explanation for the provided graph and query.

        Parameters
        ----------
        graph_data : nx.Graph
            The graph data to be analyzed.
        query : str
            The user's query to guide the explanation.

        Returns
        -------
        str
            The explanation generated by the LLM.
        """
        self.interrogator = GraphInterrogator(self.config)
        if query:
            query = sanitize_input(query)
            if not self.intent_classifier:
                self.intent_classifier = IntentClassifier()
            intent = self.intent_classifier.fit_transform(query)
            graph_info = await self.interrogator.reason_async(graph_data, query, intent)
            prompt = (
                f"Network Insights:"
                f"\nGraph Summary: {graph_info.get('Graph Summary')}\n"
                f"Descriptions of Applied Algorithms: "
                f"{graph_info.get('Descriptions of Applied Algorithms')}\n"
                f"Graph Analysis: {graph_info.get('Graph Analysis')}\n"
                f"Graph Relationships: {graph_info.get('Graph Relationships')}"
                f"\n\n**Query: {query}**\n\n"
                f"Please provide a response to the query based on the network "
                f"insights. Please avoid detailed technical descriptions unless "
                f"they help to shed light on your network-informed reasoning "
                f"mechanisms. Unless it pertains directly to the user's query, avoid "
                f"overinterpreting prominent nodes, as their hubness and "
                f"authoritativeness may merely serve to confound the analysis as a "
                f"whole and might ultimately be irrelevant to formulating a network "
                f"interpretation directed at the user's query."
            )
            logger.debug(f"Prompt:\n{prompt}")
        else:
            intent = [Intent.EXPLORATION]
            query = AnalysisPrompts.get_prompt(intent2prompt(intent))
            graph_info = await self.interrogator.reason_async(
                graph_data,
                query,
                intent,
            )
            prompt = (
                f"Network Insights:"
                f"\nGraph Summary: {graph_info.get('Graph Summary')}\n"
                f"Descriptions of Applied Algorithms: "
                f"{graph_info.get('Descriptions of Applied Algorithms')}\n"
                f"Graph Analysis: {graph_info.get('Graph Analysis')}\n"
                f"Graph Relationships: {graph_info.get('Graph Relationships')}\n\n"
                f"Please provide a comprehensive interpretation of this network."
            )
            logger.debug(f"Prompt:\n{prompt}")
        response = await self._call_llm_chain(prompt)
        return response.strip()

    def explain_algorithm(
        self, graph_data: nx.Graph, algorithm_name: str, result: Any
    ) -> str:
        """Generate a natural language explanation for a specific algorithm's results.

        Parameters
        ----------
        graph_data : networkx.Graph
            The graph on which the algorithm was run.
        algorithm_name : str
            The name of the algorithm.
        result : Any
            The result obtained from running the algorithm.

        Returns
        -------
        str
            The generated explanation.

        Raises
        ------
        Exception
            If explanation generation fails.
        """
        try:
            algorithm_docstring = inspect.getdoc(get_algorithm_function(algorithm_name))
            algorithm_description = load_algorithm_docs(algorithm_name)

            prompt = (
                f"Algorithm: {algorithm_name}\n\n"
                f"Algorithm Description:\n{algorithm_description}\n\n"
                f"Algorithm Docstring:\n{algorithm_docstring}\n\n"
                f"Result:\n{result}\n\n"
                f"Provide a clear and insightful explanation of the results of the "
                f"{algorithm_name} algorithm, considering its description and the "
                f"obtained result."
            )
            response = self.chain.run(prompt)
            return response.strip()
        except Exception:
            logger.exception(f"Error generating explanation for {algorithm_name}")
            return "Sorry, I couldn't generate an explanation at this time."

    def explain(self, graph_data: nx.Graph, query: str | None = None) -> str:
        """Generate an explanation by running the async method.

        Parameters
        ----------
        graph_data : nx.Graph
            The graph data to be analyzed.
        query : Optional[str]
            The user's query to guide the explanation.

        Returns
        -------
        str
            The explanation generated by the LLM.
        """
        return asyncio.run(self.explain_async(graph_data, query))



from nxlu._version import __version__

from .core import *
from .explanation import *
from .processing import *
from .utils import *
from .utils.log import initialize_logging

__packagename__ = "nxlu"
__url__ = "https://github.com/dpys/NxLU"

DOWNLOAD_URL = f"https://github.com/dpys/{__packagename__}/archive/{__version__}.tar.gz"

initialize_logging()



from unittest.mock import Mock, patch

import networkx as nx
import pytest
import torch
from llama_index.embeddings.huggingface import (  # Added import for assertion
    HuggingFaceEmbedding,
)
from torch import nn
from transformers import AutoModel

from nxlu.processing.embed import CustomModel, QuerySubgraph
from nxlu.processing.preprocess import create_subgraph
from nxlu.utils.control import NetworkXGraphStore


# Fixtures for various graph types with string node names
@pytest.fixture
def fully_connected_graph():
    G = nx.complete_graph(5)
    G = nx.relabel_nodes(G, lambda x: f"node{x}")  # Renaming nodes to strings
    nx.set_node_attributes(G, {n: f"Label_{n}" for n in G.nodes()}, "label")
    nx.set_edge_attributes(
        G, {edge: "relationship_type" for edge in G.edges()}, "relationship"
    )
    return G


@pytest.fixture
def partially_connected_graph():
    G = nx.Graph()
    edges = [("node0", "node1"), ("node1", "node2"), ("node3", "node4")]
    G.add_edges_from(edges)
    nx.set_node_attributes(G, {n: f"Label_{n}" for n in G.nodes()}, "label")
    nx.set_edge_attributes(
        G, {edge: "relationship_type" for edge in G.edges()}, "relationship"
    )
    return G


@pytest.fixture
def disconnected_graph():
    G = nx.Graph()
    G.add_nodes_from(["node0", "node1", "node2", "node3", "node4"])
    nx.set_node_attributes(G, {n: f"Label_{n}" for n in G.nodes()}, "label")
    return G


@pytest.fixture
def weighted_graph():
    G = nx.Graph()
    edges = [
        ("node0", "node1", 0.5),
        ("node1", "node2", 0.7),
        ("node2", "node3", 0.2),
    ]
    G.add_weighted_edges_from(edges)
    nx.set_node_attributes(G, {n: f"Label_{n}" for n in G.nodes()}, "label")
    nx.set_edge_attributes(
        G,
        {edge[:2]: "weighted_relationship" for edge in G.edges(data=True)},
        "relationship",
    )
    return G


@pytest.fixture
def binary_graph():
    G = nx.Graph()
    edges = [
        ("node0", "node1"),
        ("node1", "node2"),
        ("node2", "node3"),
        ("node3", "node4"),
    ]
    G.add_edges_from(edges)
    nx.set_node_attributes(G, {n: f"Label_{n}" for n in G.nodes()}, "label")
    nx.set_edge_attributes(
        G, {edge: "binary_relationship" for edge in G.edges()}, "relationship"
    )
    return G


@pytest.fixture(
    params=[
        "fully_connected_graph",
        "partially_connected_graph",
        "disconnected_graph",
        "weighted_graph",
        "binary_graph",
    ]
)
def varied_graph(
    request,
    fully_connected_graph,
    partially_connected_graph,
    disconnected_graph,
    weighted_graph,
    binary_graph,
):
    graph = request.getfixturevalue(request.param)
    graph.graph["name"] = request.param
    return graph


# Fixtures for CustomModel configuration
@pytest.fixture
def custom_model_config():
    return {
        "base_model": "bert-base-uncased",
        "fc_dropout": 0.3,
        "id2label": {0: "ClassA", 1: "ClassB"},
    }


# Fixtures for QuerySubgraph
@pytest.fixture
def sample_data():
    data = [
        ("node1", "node2", {"weight": 1.0, "relationship": "connected"}),
        ("node2", "node3", {"weight": 2.0, "relationship": "connected"}),
        ("node3", "node4", {"weight": 3.0, "relationship": "connected"}),
        ("node4", "node5", {"weight": 4.0, "relationship": "connected"}),
    ]
    nodes = {
        "node1": {"label": "Label_1"},
        "node2": {"label": "Label_2"},
        "node3": {"label": "Label_3"},
        "node4": {"label": "Label_4"},
        "node5": {"label": "Label_5"},
    }
    return data, nodes


# Module-level fixture for QuerySubgraph instance
@pytest.fixture
def query_subgraph_instance():
    return QuerySubgraph()


# Tests for CustomModel
class TestCustomModel:
    @patch("nxlu.processing.embed.AutoModel.from_pretrained")
    def test_custom_model_initialization(
        self, mock_from_pretrained, custom_model_config
    ):
        mock_model = Mock(spec=AutoModel)
        mock_model.config = Mock()
        mock_model.config.hidden_size = 768  # Ensure config is properly mocked
        mock_from_pretrained.return_value = mock_model

        model = CustomModel(config=custom_model_config)

        mock_from_pretrained.assert_called_once_with(custom_model_config["base_model"])
        assert model.model == mock_model
        assert isinstance(model.dropout, nn.Dropout)
        assert isinstance(model.fc, nn.Linear)
        assert model.fc.out_features == len(custom_model_config["id2label"])

    @patch("nxlu.processing.embed.AutoModel.from_pretrained")
    def test_custom_model_forward(self, mock_from_pretrained, custom_model_config):
        mock_model = Mock(spec=AutoModel)
        mock_model.config = Mock()
        mock_model.config.hidden_size = 768
        mock_model.return_value.last_hidden_state = torch.randn(2, 10, 768)
        mock_from_pretrained.return_value = mock_model

        model = CustomModel(config=custom_model_config)
        input_ids = torch.randint(0, 1000, (2, 10))
        attention_mask = torch.ones((2, 10))

        with patch.object(
            model.fc,
            "forward",
            return_value=torch.randn(2, 10, len(custom_model_config["id2label"])),
        ):
            output = model(input_ids, attention_mask)

        assert isinstance(output, torch.Tensor)
        assert output.shape == (2, len(custom_model_config["id2label"]))
        assert torch.allclose(
            output.sum(dim=1), torch.ones(2), atol=1e-5
        )  # Softmax sums to 1

    @patch("nxlu.processing.embed.AutoModel.from_pretrained")
    def test_custom_model_invalid_config(
        self, mock_from_pretrained, custom_model_config
    ):
        mock_model = Mock(spec=AutoModel)
        mock_model.config = Mock()
        mock_model.config.hidden_size = 768  # Ensure config is properly mocked
        mock_from_pretrained.return_value = mock_model

        invalid_config = custom_model_config.copy()
        invalid_config["id2label"] = {}  # Empty id2label

        model = CustomModel(config=invalid_config)

        mock_from_pretrained.assert_called_once_with(custom_model_config["base_model"])
        assert model.model == mock_model
        assert isinstance(model.dropout, nn.Dropout)
        assert isinstance(model.fc, nn.Linear)
        assert model.fc.out_features == 0  # Updated assertion


class TestQuerySubgraph:
    def test_query_subgraph_initialization(self, query_subgraph_instance):
        assert isinstance(query_subgraph_instance.data_graph, nx.Graph)
        assert isinstance(query_subgraph_instance.graph_store, NetworkXGraphStore)
        assert query_subgraph_instance.index_nodes is None
        assert query_subgraph_instance.index_edges is None
        assert isinstance(query_subgraph_instance.embedding_model, HuggingFaceEmbedding)

    @patch("nxlu.processing.embed.create_subgraph")
    def test_load_data_graph(
        self, mock_create_subgraph, query_subgraph_instance, sample_data
    ):
        data, nodes = sample_data
        query_subgraph_instance.load_data_graph(data=data, nodes=nodes)

        # Ensure token nodes are excluded
        # Since none are token nodes in sample_data, all should be added
        assert query_subgraph_instance.data_graph.number_of_nodes() == 5
        assert query_subgraph_instance.data_graph.number_of_edges() == 4
        for node, attrs in nodes.items():
            assert (
                query_subgraph_instance.data_graph.nodes[node]["label"]
                == attrs["label"]
            )

    @patch("nxlu.processing.embed.create_subgraph")
    def test_load_data_graph_with_token_nodes(
        self, mock_create_subgraph, query_subgraph_instance, sample_data
    ):
        data, nodes = sample_data
        # Add token nodes
        nodes["token1"] = {"type": "token", "label": "Token_1"}
        nodes["node6"] = {"label": "Label_6"}

        data += [("node1", "token1", {"weight": 0.5, "relationship": "token_relation"})]
        query_subgraph_instance.load_data_graph(data=data, nodes=nodes)

        # Token nodes should be excluded
        assert (
            query_subgraph_instance.data_graph.number_of_nodes() == 5
        )  # node1 to node5 and node6
        assert query_subgraph_instance.data_graph.has_node("token1") is False
        assert query_subgraph_instance.data_graph.has_edge("node1", "token1") is False

    @patch("nxlu.processing.embed.VectorStoreIndex")
    @patch("nxlu.processing.embed.StorageContext.from_defaults")
    @patch("nxlu.processing.embed.create_subgraph")
    def test_prepare_node_index(
        self,
        mock_create_subgraph,
        mock_storage_context,
        mock_vector_store_index,
        query_subgraph_instance,
        sample_data,
    ):
        data, nodes = sample_data
        query_subgraph_instance.load_data_graph(data=data, nodes=nodes)

        # Mock VectorStoreIndex
        mock_index = Mock()
        mock_vector_store_index.return_value = mock_index

        query_subgraph_instance.prepare_node_index()

        mock_storage_context.assert_called_once()
        mock_vector_store_index.assert_called_once()
        assert query_subgraph_instance.index_nodes == mock_index

    @patch("nxlu.processing.embed.VectorStoreIndex")
    @patch("nxlu.processing.embed.StorageContext.from_defaults")
    @patch("nxlu.processing.embed.create_subgraph")
    def test_prepare_edge_index(
        self,
        mock_create_subgraph,
        mock_storage_context,
        mock_vector_store_index,
        query_subgraph_instance,
        sample_data,
    ):
        data, nodes = sample_data
        query_subgraph_instance.load_data_graph(data=data, nodes=nodes)

        # Mock VectorStoreIndex
        mock_index = Mock()
        mock_vector_store_index.return_value = mock_index

        subgraph = query_subgraph_instance.data_graph.copy()
        query_subgraph_instance.prepare_edge_index(subgraph=subgraph)

        mock_storage_context.assert_called_once()
        mock_vector_store_index.assert_called_once()
        assert query_subgraph_instance.index_edges == mock_index

    @patch("nxlu.processing.embed.VectorStoreIndex")
    @patch("nxlu.processing.embed.StorageContext.from_defaults")
    def test_query_graph(
        self,
        mock_storage_context,
        mock_vector_store_index,
        query_subgraph_instance,
        sample_data,
    ):
        data, nodes = sample_data
        query_subgraph_instance.load_data_graph(data=data, nodes=nodes)

        # Create a mock VectorStoreIndex instance
        mock_index_nodes = Mock()
        mock_vector_store_index.return_value = mock_index_nodes

        # Create a mock QueryEngine
        mock_query_engine = Mock()

        # Create a mock source node with correctly formatted text
        mock_source_node = Mock()
        mock_source_node.text = "Node: node1, Attributes: {'label': 'Label_1'}"

        # Mock the response of the query
        mock_query_response = Mock()
        mock_query_response.source_nodes = [mock_source_node]

        # Configure the mock_query_engine to return the mock_query_response
        mock_query_engine.query.return_value = mock_query_response

        # Configure the mock_index_nodes to return the mock_query_engine when
        # as_query_engine is called
        mock_index_nodes.as_query_engine.return_value = mock_query_engine

        # Now call the method under test
        node_ids, edge_tuples = query_subgraph_instance.query_graph(query="test query")

        # Check that the node_ids are as expected
        assert node_ids == ["node1"]
        # Since no edges are mocked, edge_tuples should be empty
        assert edge_tuples == []


class TestQuerySubgraphEdgeCases:
    @pytest.fixture
    def empty_graph(self):
        G = nx.Graph()
        return G

    @patch("nxlu.processing.embed.QuerySubgraph.query_graph")
    @patch("nxlu.processing.embed.QuerySubgraph.prepare_node_index")
    @patch("nxlu.processing.embed.QuerySubgraph.prepare_edge_index")
    def test_query_graph_with_empty_graph(
        self,
        mock_prepare_edge_index,
        mock_prepare_node_index,
        mock_query_graph,
        query_subgraph_instance,
        empty_graph,
    ):
        # Mock the query response to avoid calling the actual LLM or OpenAI API
        mock_query_graph.return_value = ([], [])

        query_subgraph_instance.load_data_graph(data=[], nodes={})
        node_ids, edge_tuples = query_subgraph_instance.query_graph(query="test query")

        assert node_ids == []
        assert edge_tuples == []

    @patch("nxlu.processing.embed.QuerySubgraph.query_graph")
    @patch("nxlu.processing.embed.QuerySubgraph.prepare_node_index")
    @patch("nxlu.processing.embed.QuerySubgraph.prepare_edge_index")
    def test_query_graph_with_no_matching_nodes(
        self,
        mock_prepare_edge_index,
        mock_prepare_node_index,
        mock_query_graph,
        query_subgraph_instance,
        sample_data,
    ):
        data, nodes = sample_data
        query_subgraph_instance.load_data_graph(data=data, nodes=nodes)

        # Mock query_graph to return nodes that do not exist
        mock_query_graph.return_value = (["nonexistent"], [])

        node_ids, edge_tuples = query_subgraph_instance.query_graph(query="no match")

        assert node_ids == ["nonexistent"]
        assert edge_tuples == []
        # create_subgraph should handle non-existent nodes gracefully
        with patch("nxlu.processing.embed.create_subgraph", return_value=nx.Graph()):
            subgraph = query_subgraph_instance.create_query_subgraph(
                graph=query_subgraph_instance.data_graph,
                query="no match",
                top_k_nodes=1,
                top_k_edges=0,
            )
            assert isinstance(subgraph, nx.Graph)
            assert subgraph.number_of_nodes() == 0
            assert subgraph.number_of_edges() == 0

    @patch("nxlu.processing.embed.QuerySubgraph.query_graph")
    @patch("nxlu.processing.embed.QuerySubgraph.prepare_node_index")
    @patch("nxlu.processing.embed.QuerySubgraph.prepare_edge_index")
    def test_query_graph_with_all_token_nodes(
        self,
        mock_prepare_edge_index,
        mock_prepare_node_index,
        mock_query_graph,
        query_subgraph_instance,
        sample_data,
    ):
        data, nodes = sample_data
        # All nodes are token nodes
        nodes = {node: {"type": "token"} for node in nodes}
        query_subgraph_instance.load_data_graph(data=data, nodes=nodes)

        # Mock query_graph to return empty node_ids and edge_tuples
        mock_query_graph.return_value = ([], [])

        node_ids, edge_tuples = query_subgraph_instance.query_graph(query="test query")
        # Since all nodes are token nodes, node_ids should be empty after filtering
        assert node_ids == []
        assert edge_tuples == []

    @patch("nxlu.processing.embed.QuerySubgraph.query_graph")
    def test_query_graph_with_large_k_values(
        self,
        mock_query_graph,
        query_subgraph_instance,
        sample_data,
    ):
        data, nodes = sample_data
        query_subgraph_instance.load_data_graph(data=data, nodes=nodes)

        # Mock query_graph to return all nodes and edges
        mock_query_graph.return_value = (
            ["node1", "node2", "node3", "node4", "node5"],
            [
                ("node1", "node2"),
                ("node2", "node3"),
                ("node3", "node4"),
                ("node4", "node5"),
            ],
        )

        # Mock create_subgraph to return the expected subgraph
        subgraph = create_subgraph(
            query_subgraph_instance.data_graph,
            ["node1", "node2", "node3", "node4", "node5"],
            [
                ("node1", "node2"),
                ("node2", "node3"),
                ("node3", "node4"),
                ("node4", "node5"),
            ],
        )
        with patch("nxlu.processing.embed.create_subgraph", return_value=subgraph):
            result_subgraph = query_subgraph_instance.create_query_subgraph(
                graph=query_subgraph_instance.data_graph,
                query="test query",
                top_k_nodes=10,  # Larger than available nodes
                top_k_edges=10,  # Larger than available edges
            )

            assert isinstance(result_subgraph, nx.Graph)
            assert result_subgraph.number_of_nodes() == 5
            assert result_subgraph.number_of_edges() == 4

    @patch("nxlu.processing.embed.QuerySubgraph._create_ego_subgraph")
    @patch("nxlu.processing.embed.QuerySubgraph.query_graph")
    @patch("nxlu.processing.embed.QuerySubgraph.prepare_node_index")
    @patch("nxlu.processing.embed.QuerySubgraph.prepare_edge_index")
    def test_query_graph_with_duplicate_edges(
        self,
        mock_prepare_edge_index,
        mock_prepare_node_index,
        mock_query_graph,
        mock_create_ego_subgraph,
        query_subgraph_instance,
        sample_data,
    ):
        data, nodes = sample_data
        # Add duplicate edges
        data += [("node1", "node2", {"weight": 1.0, "relationship": "connected"})]
        query_subgraph_instance.load_data_graph(data=data, nodes=nodes)

        # Mock the response of query_graph to return duplicate edges
        mock_query_graph.return_value = (
            ["node1", "node2"],
            [("node1", "node2"), ("node1", "node2")],  # Duplicate edges
        )

        # Create a mocked DiGraph subgraph with 2 nodes and 1 edge
        mocked_subgraph = nx.DiGraph()
        mocked_subgraph.add_nodes_from(["node1", "node2"])
        mocked_subgraph.add_edge("node1", "node2", weight=1.0, relationship="connected")

        # Configure the mock to return the mocked DiGraph
        mock_create_ego_subgraph.return_value = mocked_subgraph

        # Call the method under test
        result_subgraph = query_subgraph_instance.create_query_subgraph(
            graph=query_subgraph_instance.data_graph,
            query="test query",
            top_k_nodes=2,
            top_k_edges=2,
        )

        # Assertions
        assert isinstance(result_subgraph, nx.DiGraph), "Subgraph should be a DiGraph."
        assert result_subgraph.number_of_nodes() == 2, "Subgraph should have 2 nodes."
        assert (
            result_subgraph.number_of_edges() == 1
        ), "Subgraph should have 1 edge without duplicates."
        assert result_subgraph.has_edge(
            "node1", "node2"
        ), "Edge (node1, node2) should exist."


class TestCustomModelEdgeCases:
    @patch("nxlu.processing.embed.AutoModel.from_pretrained")
    def test_custom_model_zero_classes(self, mock_from_pretrained, custom_model_config):
        mock_model = Mock(spec=AutoModel)
        mock_model.config = Mock()
        mock_model.config.hidden_size = 768
        mock_from_pretrained.return_value = mock_model

        invalid_config = custom_model_config.copy()
        invalid_config["id2label"] = {}

        model = CustomModel(config=invalid_config)

        mock_from_pretrained.assert_called_once_with(custom_model_config["base_model"])
        assert model.fc.out_features == 0  # Updated assertion

    @patch("nxlu.processing.embed.AutoModel.from_pretrained")
    def test_custom_model_forward_incorrect_input_shape(
        self, mock_from_pretrained, custom_model_config
    ):
        mock_model = Mock(spec=AutoModel)
        mock_model.config = Mock()
        mock_model.config.hidden_size = 768
        # Mock the model to return a last_hidden_state with variable sequence length
        mock_model.return_value.last_hidden_state = torch.randn(
            2, 8, 768
        )  # Changed sequence length
        mock_from_pretrained.return_value = mock_model

        model = CustomModel(config=custom_model_config)
        # Different input dimensions
        input_ids = torch.randint(0, 1000, (2, 8))
        attention_mask = torch.ones((2, 8))

        with patch.object(
            model.fc,
            "forward",
            return_value=torch.randn(2, 8, len(custom_model_config["id2label"])),
        ):
            output = model(input_ids, attention_mask)

        assert isinstance(output, torch.Tensor)
        assert output.shape == (2, len(custom_model_config["id2label"]))
        assert torch.allclose(
            output.sum(dim=1), torch.ones(2), atol=1e-5
        )  # Softmax sums to 1


class TestQuerySubgraphIntegration:
    @pytest.fixture
    def complex_graph(self):
        G = nx.Graph()
        # Add nodes
        for i in range(10):
            G.add_node(f"node{i}", label=f"Label_{i}")
        # Add edges with weights and relationships
        edges = [
            ("node0", "node1", {"weight": 1.0, "relationship": "connected"}),
            ("node1", "node2", {"weight": 2.0, "relationship": "connected"}),
            ("node2", "node3", {"weight": 3.0, "relationship": "connected"}),
            ("node3", "node4", {"weight": 4.0, "relationship": "connected"}),
            ("node4", "node5", {"weight": 5.0, "relationship": "connected"}),
            ("node5", "node6", {"weight": 6.0, "relationship": "connected"}),
            ("node6", "node7", {"weight": 7.0, "relationship": "connected"}),
            ("node7", "node8", {"weight": 8.0, "relationship": "connected"}),
            ("node8", "node9", {"weight": 9.0, "relationship": "connected"}),
            ("node9", "node0", {"weight": 10.0, "relationship": "connected"}),
            # Token nodes
            ("node0", "token1", {"weight": 0.1, "relationship": "token_relation"}),
            ("token1", "token2", {"weight": 0.2, "relationship": "token_relation"}),
        ]
        G.add_edges_from(edges)
        return G

    @patch("nxlu.processing.embed.QuerySubgraph._create_ego_subgraph")
    def test_create_query_subgraph_complex(
        self,
        mock_create_ego_subgraph,
        query_subgraph_instance,
        complex_graph,
    ):
        query_subgraph_instance.load_data_graph(
            data=list(complex_graph.edges(data=True)),
            nodes=dict(complex_graph.nodes(data=True)),
        )

        # Mock the query_graph to return specific nodes and edges
        with patch.object(
            query_subgraph_instance,
            "query_graph",
            return_value=(
                ["node1", "node2", "node3"],
                [("node1", "node2"), ("node2", "node3")],
            ),
        ):
            # Create a mocked DiGraph subgraph with 3 nodes and 2 edges
            mocked_subgraph = nx.DiGraph()
            mocked_subgraph.add_nodes_from(["node1", "node2", "node3"])
            mocked_subgraph.add_edges_from([("node1", "node2"), ("node2", "node3")])

            # Configure the mock to return the mocked DiGraph
            mock_create_ego_subgraph.return_value = mocked_subgraph

            # Call the method under test
            subgraph_result = query_subgraph_instance.create_query_subgraph(
                graph=complex_graph, query="connectivity", top_k_nodes=3, top_k_edges=2
            )

            # Assertions
            assert isinstance(
                subgraph_result, nx.DiGraph
            ), "Subgraph should be a DiGraph."
            assert (
                subgraph_result.number_of_nodes() == 3
            ), "Subgraph should have 3 nodes."
            assert (
                subgraph_result.number_of_edges() == 2
            ), "Subgraph should have 2 edges."
            assert subgraph_result.has_edge(
                "node1", "node2"
            ), "Edge (node1, node2) should exist."
            assert subgraph_result.has_edge(
                "node2", "node3"
            ), "Edge (node2, node3) should exist."

    def test_create_subgraph_with_no_edges(self, varied_graph):
        node_subset = ["node1", "node2"]
        edge_subset = []
        subgraph = QuerySubgraph.slice_subgraph(varied_graph, node_subset, edge_subset)

        assert isinstance(subgraph, nx.Graph)
        assert subgraph.number_of_nodes() == 2

        # Dynamically determine the expected number of edges
        if varied_graph.has_edge("node1", "node2"):
            expected_edges = 1
        else:
            expected_edges = 0

        assert subgraph.number_of_edges() == expected_edges  # Updated assertion

    @patch("nxlu.processing.embed.QuerySubgraph._create_ego_subgraph")
    def test_create_query_subgraph_disconnected_subgraph(
        self, mock_create_ego_subgraph, query_subgraph_instance, sample_data
    ):
        data, nodes = sample_data
        query_subgraph_instance.load_data_graph(data=data, nodes=nodes)

        with patch.object(
            query_subgraph_instance,
            "query_graph",
            return_value=(["node1", "node3"], []),
        ):
            mocked_subgraph = create_subgraph(
                query_subgraph_instance.data_graph, ["node1", "node3"], []
            )
            mock_create_ego_subgraph.return_value = mocked_subgraph

            result_subgraph = query_subgraph_instance.create_query_subgraph(
                graph=query_subgraph_instance.data_graph,
                query="test query",
                top_k_nodes=2,
                top_k_edges=0,
            )

            assert isinstance(result_subgraph, nx.Graph)
            assert result_subgraph.number_of_nodes() == 2
            assert result_subgraph.number_of_edges() == 0
            assert not nx.is_weakly_connected(result_subgraph)



import logging
import warnings
from typing import Any

import networkx as nx
import torch
from huggingface_hub import PyTorchModelHubMixin
from llama_index.core import StorageContext, VectorStoreIndex
from llama_index.core.schema import TextNode
from llama_index.embeddings.huggingface import HuggingFaceEmbedding
from torch import nn
from transformers import AutoModel

from nxlu.processing.preprocess import create_subgraph, lcc
from nxlu.utils.control import NetworkXGraphStore

warnings.filterwarnings("ignore")


logger = logging.getLogger("nxlu")


__all__ = ["CustomModel", "QuerySubgraph"]


class CustomModel(nn.Module, PyTorchModelHubMixin):
    """Custom neural network model for domain classification.

    Attributes
    ----------
    model : transformers.AutoModel
        The pre-trained transformer model.
    dropout : torch.nn.Dropout
        Dropout layer for regularization.
    fc : torch.nn.Linear
        Fully connected layer for classification.
    """

    def __init__(self, config):
        super().__init__()
        self.model = AutoModel.from_pretrained(config["base_model"])
        self.dropout = nn.Dropout(config["fc_dropout"])
        self.fc = nn.Linear(self.model.config.hidden_size, len(config["id2label"]))

    def forward(self, input_ids, attention_mask):
        """Forward pass through the network.

        Parameters
        ----------
        input_ids : torch.Tensor
            Token IDs.
        attention_mask : torch.Tensor
            Attention masks.

        Returns
        -------
        torch.Tensor
            Softmax probabilities for each class.
        """
        features = self.model(
            input_ids=input_ids, attention_mask=attention_mask
        ).last_hidden_state
        dropped = self.dropout(features)
        outputs = self.fc(dropped)
        return torch.softmax(outputs[:, 0, :], dim=1)


class QuerySubgraph:
    """A class to manage and query a subgraph using vector embeddings.

    It maintains separate node and edge indices for comprehensive querying.

    Attributes
    ----------
    data_graph : nx.DiGraph
        The NetworkX directed graph containing the data to be queried.
    graph_store : NetworkXGraphStore or None
        A store for managing the graph structure and data.
    index_nodes : VectorStoreIndex or None
        The vector store index for querying nodes in the graph.
    index_edges : VectorStoreIndex or None
        The vector store index for querying edges in the graph.
    embedding_model : HuggingFaceEmbedding
        The embedding model used for indexing and querying nodes and edges.
    similarity_metric : str
        The similarity metric used for querying (e.g., 'cosine').
    mmr_threshold : float
        The threshold for Maximal Marginal Relevance (MMR).
    """

    def __init__(
        self,
        similarity_metric: str = "cosine",
        mmr_threshold: float = 0.2,
        embedding_model_name: str = "all-MiniLM-L6-v2",
    ):
        """Initialize the QuerySubgraph class with an empty directed graph, a graph
        store, separate indices for nodes and edges, and an embedding model for
        vector-based queries. Includes caching capabilities to speed up repeated
        queries.

        Parameters
        ----------
        similarity_metric : str
            The similarity metric used for querying (e.g., 'cosine').
        mmr_threshold : float
            The threshold for Maximal Marginal Relevance (MMR).
        embedding_model_name : str
            The name of the pre-trained embedding model.
        """
        self.data_graph = nx.DiGraph()  # Use DiGraph for directed graphs
        self.graph_store = NetworkXGraphStore(self.data_graph)
        self.similarity_metric = similarity_metric
        self.mmr_threshold = mmr_threshold
        self.index_nodes = None
        self.index_edges = None
        self.embedding_model = HuggingFaceEmbedding(model_name=embedding_model_name)

    def load_data_graph(
        self,
        data: list[tuple[str, str, dict[str, Any]]],
        nodes: dict[str, dict[str, Any]],
        min_component_size: int = 2,  # Minimum size to consider a component significant
    ):
        """Load the actual data graph, ensuring only nodes in significant weakly
        connected components are included.

        Parameters
        ----------
        data : List[Tuple[str, str, Dict[str, Any]]]
            A list of edges in the form of (node1, node2, attributes) where attributes
            contain edge information (e.g., weight).
        nodes : Dict[str, Dict[str, Any]]
            A dictionary where keys are node identifiers and values are dictionaries of
            node attributes.
        min_component_size : int, optional
            The minimum number of nodes a connected component must have to be included,
            by default 2.

        Returns
        -------
        None
        """
        if not isinstance(data, list) or not all(
            isinstance(edge, tuple) and len(edge) == 3 for edge in data
        ):
            logger.error("Invalid data format for edges.")
            return
        if not isinstance(nodes, dict):
            logger.error("Invalid data format for nodes.")
            return

        try:
            # First, add all nodes that are not token nodes
            for node, attrs in nodes.items():
                node_str = str(node)
                if not self._is_token_node(node_str, attrs):
                    self.data_graph.add_node(node_str, **attrs)
                    logger.debug(f"Added node: {node_str} with attributes: {attrs}")

            # Add all edges where both nodes are not token nodes
            filtered_data = [
                (str(u), str(v), d)
                for u, v, d in data
                if not self._is_token_node(str(u), nodes.get(u, {}))
                and not self._is_token_node(str(v), nodes.get(v, {}))
            ]

            if not filtered_data:
                logger.warning("No valid edges to add after filtering token nodes.")

            # Add weighted edges
            self.data_graph.add_weighted_edges_from(
                [(u, v, d.get("weight", 1.0)) for u, v, d in filtered_data]
            )
            logger.info(
                f"Data graph loaded with {self.data_graph.number_of_nodes()} nodes and "
                f"{self.data_graph.number_of_edges()} edges."
            )

            # Identify weakly connected components
            components = list(nx.weakly_connected_components(self.data_graph))
            logger.info(f"Identified {len(components)} weakly connected components.")

            # Filter components based on size
            significant_components = [
                comp for comp in components if len(comp) >= min_component_size
            ]
            logger.info(
                f"{len(significant_components)} components meet the size threshold of "
                f"{min_component_size}."
            )

            if not significant_components:
                logger.warning(
                    "No connected components meet the minimum size threshold."
                )
                self.data_graph = nx.DiGraph()
                self.graph_store = NetworkXGraphStore(self.data_graph)
                return

            # Create a new graph with only significant components
            filtered_graph = nx.DiGraph()
            for comp in significant_components:
                filtered_graph.add_nodes_from(
                    (node, self.data_graph.nodes[node]) for node in comp
                )
                filtered_graph.add_edges_from(
                    (u, v, self.data_graph.edges[u, v])
                    for u, v in self.data_graph.edges
                    if u in comp and v in comp
                )
                logger.debug(f"Added component with {len(comp)} nodes.")

            # Replace the original graph with the filtered graph
            self.data_graph = filtered_graph
            self.graph_store = NetworkXGraphStore(self.data_graph)
            logger.info(
                f"Filtered data graph now has {self.data_graph.number_of_nodes()} "
                f"nodes and {self.data_graph.number_of_edges()} edges."
            )
        except Exception:
            logger.exception("Failed to load and filter data graph")
            raise

    def _is_token_node(self, node: str, attrs: dict[str, Any]) -> bool:
        """Determine if a node is a token node based on attributes.

        Parameters
        ----------
        node : str
            The node identifier.
        attrs : Dict[str, Any]
            The attributes associated with the node.

        Returns
        -------
        bool
            True if the node is identified as a token node, False otherwise.
        """
        return attrs.get("type") == "token" if attrs else False

    def _extract_node_id(self, node_text: str) -> str | None:
        """
        Extract the node ID from the node text.
        Assumes the format: "Node: {node}, Attributes: {data}"

        Parameters
        ----------
        node_text : str
            The text representation of the node.

        Returns
        -------
        Optional[str]
            The extracted node ID, or None if extraction fails.
        """
        try:
            prefix = "Node:"
            comma_index = node_text.find(",")
            if comma_index == -1:
                logger.warning(f"Comma not found in node text: {node_text}")
                return None
            node_part = node_text[:comma_index]
            if not node_part.startswith(prefix):
                logger.warning(f"Node text does not start with '{prefix}': {node_text}")
                return None
            node_id = node_part[len(prefix) :].strip()
        except Exception:
            logger.exception(f"Error extracting node ID from text '{node_text}'")
            return None
        else:
            return node_id

    def _extract_edge_tuple(self, edge_text: str) -> tuple[str, str] | None:
        """Extract the edge tuple (u, v) or (u, v, key) from the edge text.

        Assumes: "Edge: {u} -- {relation} (Weight: {weight}) --> {v} | ID: {edge_id}"

        Parameters
        ----------
        edge_text : str
            The text representation of the edge.

        Returns
        -------
        Optional[Tuple[str, str]]
            The extracted edge tuple, or None if extraction fails.
        """
        try:
            edge_id = edge_text.split("| ID:")[1].strip()
            return tuple(edge_id.split("-"))
        except IndexError:
            logger.warning(f"Unable to extract edge tuple from text: {edge_text}")
            return None
        except Exception:
            logger.exception(f"Error extracting edge tuple from text '{edge_text}'")
            return None

    def prepare_node_index(self):
        """Prepare the VectorStoreIndex for nodes by indexing the data graph's nodes.
        Utilizes caching to avoid recomputing indices.

        Returns
        -------
        None
        """
        logger.info("Preparing node index for efficient querying.")

        storage_context_nodes = StorageContext.from_defaults(
            graph_store=self.graph_store
        )

        node_texts = [
            TextNode(text=f"Node: {node}, Attributes: {data}", id_=str(node))
            for node, data in self.data_graph.nodes(data=True)
            if not self._is_token_node(node, data)
        ]

        self.index_nodes = VectorStoreIndex(
            nodes=node_texts,
            storage_context=storage_context_nodes,
            embed_model=self.embedding_model,
        )

    def prepare_edge_index(self, subgraph: nx.Graph):
        """Prepare the VectorStoreIndex for edges by indexing the edges of a subgraph.
        Utilizes caching to avoid recomputing indices.

        Parameters
        ----------
        subgraph : nx.Graph or nx.MultiGraph
            The subgraph containing the nodes and edges to be indexed.

        Returns
        -------
        None
        """
        logger.info("Preparing edge index for efficient querying.")

        if isinstance(subgraph, (nx.MultiGraph, nx.MultiDiGraph)):
            edge_texts = [
                TextNode(
                    text=f"Edge: {u} -- {data.get('relation', 'EDGE')} "
                    f"(Weight: {data.get('weight', 'N/A')}) --> {v} | ID: "
                    f"{u}-{v}-{key}",
                    id_=f"{u}-{v}-{key}",
                )
                for u, v, key, data in subgraph.edges(keys=True, data=True)
                if not self._is_token_node(u, subgraph.nodes[u])
                and not self._is_token_node(v, subgraph.nodes[v])
            ]
        else:
            edge_texts = [
                TextNode(
                    text=f"Edge: {u} -- {data.get('relation', 'EDGE')} "
                    f"(Weight: {data.get('weight', 'N/A')}) --> {v} | ID: {u}-{v}",
                    id_=f"{u}-{v}",
                )
                for u, v, data in subgraph.edges(data=True)
                if not self._is_token_node(u, subgraph.nodes[u])
                and not self._is_token_node(v, subgraph.nodes[v])
            ]

        self.index_edges = VectorStoreIndex(
            nodes=edge_texts,
            storage_context=StorageContext.from_defaults(graph_store=self.graph_store),
            embed_model=self.embedding_model,
        )

    def query_graph(
        self,
        query: str,
        top_k_nodes: int = 1000,
        top_k_edges: int = 1000000,
    ) -> tuple[list[str], list[tuple[str, str]]]:
        """Query both the node and edge indices to retrieve relevant nodes and edges.

        Parameters
        ----------
        query : str
            The user's query to search for relevant nodes and edges.
        top_k_nodes : int, optional
            A count of the top relevant nodes to retrieve, by default 1000.
        top_k_edges : int, optional
            A count of the top relevant edges to retrieve, by default 1000000.

        Returns
        -------
        Tuple[List[str], List[Tuple[str, str]]]
            A tuple containing two lists:
            - List of node IDs that are most relevant to the query.
            - List of edge tuples that are most relevant to the query, constrained to
              the relevant nodes.
        """
        logger.info(f"Querying graph with: {query}")

        self.prepare_node_index()

        if not self.index_nodes:
            logger.error(
                "Node index has not been prepared. Call prepare_node_index first."
            )
            return [], []

        node_query_engine = self.index_nodes.as_query_engine(
            similarity_top_k=top_k_nodes,
            mmr_threshold=self.mmr_threshold,
            similarity_metric=self.similarity_metric,
        )
        node_response = node_query_engine.query(query)
        relevant_nodes_text = [node.text for node in node_response.source_nodes]
        logger.debug(f"Relevant nodes retrieved: {relevant_nodes_text}")

        if not relevant_nodes_text:
            logger.warning(
                "No relevant nodes found. Extracting the largest connected component."
            )

            lcc_graph = lcc(self.data_graph)
            logger.info(
                f"Largest connected component has {lcc_graph.number_of_nodes()} nodes."
            )

            if lcc_graph.number_of_nodes() == 0:
                logger.warning("No connected components found in the graph.")
                return [], []

            node_ids = [str(node) for node in lcc_graph.nodes()]

            node_filtered_subgraph = lcc_graph

            self.prepare_edge_index(node_filtered_subgraph)

            if not self.index_edges:
                logger.error(
                    "Edge index has not been prepared. Call prepare_edge_index first."
                )
                return node_ids, []

            edge_tuples = list(node_filtered_subgraph.edges())
            return node_ids, edge_tuples

        node_ids = [
            self._extract_node_id(node_text) for node_text in relevant_nodes_text
        ]
        node_ids = [nid for nid in node_ids if nid and self.data_graph.has_node(nid)]
        logger.debug(f"Filtered node IDs present in data graph: {node_ids}")

        node_filtered_subgraph = self.slice_subgraph(self.data_graph, node_ids)

        self.prepare_edge_index(node_filtered_subgraph)

        if not self.index_edges:
            logger.error(
                "Edge index has not been prepared. Call prepare_edge_index first."
            )
            return node_ids, []

        edge_query_engine = self.index_edges.as_query_engine(
            similarity_top_k=top_k_edges
        )
        edge_response = edge_query_engine.query(query)
        relevant_edges_text = [edge.text for edge in edge_response.source_nodes]
        logger.debug(f"Relevant edges retrieved: {relevant_edges_text}")

        edge_tuples = [
            self._extract_edge_tuple(edge_text) for edge_text in relevant_edges_text
        ]
        edge_tuples = [
            et for et in edge_tuples if et and self.data_graph.has_edge(et[0], et[1])
        ]
        logger.debug(f"Extracted and filtered edge tuples: {edge_tuples}")

        return node_ids, edge_tuples

    @classmethod
    def slice_subgraph(
        cls,
        graph_data: nx.Graph,
        node_subset: list[str],
        edge_subset: list[tuple[str, str]] | None = None,
    ) -> nx.Graph:
        return create_subgraph(graph_data, node_subset, edge_subset)

    def _determine_optimal_radius(self, node_ids: list[str]) -> int:
        """Determine the optimal radius for ego graph based on the relevance of nodes.

        Parameters
        ----------
        node_ids : List[str]
            List of node IDs that are most relevant to the query.

        Returns
        -------
        int
            The optimal radius value.
        """
        max_radius = 5
        target_size = 50

        for radius in range(1, max_radius + 1):
            ego_subgraph = self._create_ego_subgraph(node_ids, radius)
            num_nodes = ego_subgraph.number_of_nodes()
            logger.debug(f"Radius {radius}: Subgraph has {num_nodes} nodes.")
            if num_nodes >= target_size:
                return radius
        return max_radius

    def _create_ego_subgraph(self, node_ids: list[str], radius: int) -> nx.Graph:
        """Create an ego graph subgraph for the given nodes and radius.

        Parameters
        ----------
        node_ids : List[str]
            List of central node IDs.
        radius : int
            The radius for the ego graph.

        Returns
        -------
        nx.Graph
            The combined ego subgraph.
        """
        ego_graphs = []
        for node_id in node_ids:
            if self.data_graph.has_node(node_id):
                ego_graph = nx.ego_graph(
                    self.data_graph, node_id, radius=radius, undirected=False
                )
                if ego_graph.number_of_nodes() > 0:
                    ego_graphs.append(ego_graph)
                    logger.debug(
                        f"Ego graph for node {node_id} with radius {radius} has "
                        f"{ego_graph.number_of_nodes()} nodes."
                    )
            else:
                logger.warning(f"Node {node_id} not found in the data graph.")

        if not ego_graphs:
            logger.warning("No ego graphs were created. Returning an empty graph.")
            return nx.DiGraph()

        combined_subgraph = nx.compose_all(ego_graphs)
        return combined_subgraph

    def create_query_subgraph(
        self,
        graph: nx.Graph,
        query: str,
        top_k_nodes: int = 1000,
        top_k_edges: int = 1000000,
    ) -> nx.Graph:
        """Create a subgraph based on the relevance to the given query.

        Parameters
        ----------
        graph : nx.Graph
            The complete graph to query from.
        query : str
            The query string to search for relevant nodes and edges.
        top_k_nodes : int, optional
            Number of top relevant nodes to retrieve, by default 1000.
        top_k_edges : int, optional
            Number of top relevant edges to retrieve, by default 1000000.

        Returns
        -------
        nx.Graph
            The resulting subgraph containing the most relevant nodes and edges.
        """
        try:
            self.load_data_graph(
                data=list(graph.edges(data=True)),
                nodes=dict(graph.nodes(data=True)),
            )
            logger.info("Data graph loaded into QuerySubgraph.")
            logger.debug(f"Data graph nodes: {list(self.data_graph.nodes())}")
        except Exception:
            logger.exception("Error loading data graph into QuerySubgraph.")
            raise

        try:
            node_ids, _ = self.query_graph(
                query=query,
                top_k_nodes=top_k_nodes,
                top_k_edges=top_k_edges,
            )
            logger.debug(f"Node IDs after query and filtering: {node_ids}")

            if not node_ids:
                logger.warning("No relevant nodes found for the query.")
                return nx.DiGraph()  # Return empty graph

            optimal_radius = self._determine_optimal_radius(node_ids)
            logger.info(f"Optimal radius determined: {optimal_radius}")

            subgraph = self._create_ego_subgraph(node_ids, optimal_radius)
            logger.info(
                f"Ego subgraph created using radius {optimal_radius}, resulting in "
                f"{subgraph.number_of_nodes()} nodes and {subgraph.number_of_edges()} "
                f"edges."
            )
            if subgraph.number_of_nodes() < 3 or not nx.is_weakly_connected(subgraph):
                logger.warning(
                    "Subgraph is too small or not weakly connected after applying ego "
                    "graph. Ensure that your query is relevant to the network that "
                    "you are querying."
                )
        except Exception:
            logger.exception("Error during querying or creating ego subgraph")
            raise

        return subgraph



from enum import Enum

__all__ = ["Framework", "OpenAIModel", "AnthropicModel", "LocalModel"]


class Framework(str, Enum):
    LANGCHAIN = "langchain"
    LLAMAINDEX = "llamaindex"


class OpenAIModel(str, Enum):
    GPT_4_TURBO = "gpt-4-turbo"
    GPT_4 = "gpt-4"
    GPT_4O = "gpt-4o-2024-08-06"
    GPT_4O_MINI = "gpt-4o-mini"
    GPT_4O1_PREVIEW = "o1-preview"
    GPT_401_MINI = "o1-mini"


class AnthropicModel(str, Enum):
    CLAUDE_2 = "claude-2"
    CLAUDE_2_0 = "claude-2.0"
    CLAUDE_INSTANT = "claude-instant"
    CLAUDE_INSTANT_1 = "claude-instant-1"
    CLAUDE_INSTANT_1_1 = "claude-instant-1.1"
    CLAUDE_3_5_SONNET = "claude-3.5-sonnet"
    CLAUDE_3_SONNET = "claude-3-sonnet"


class LocalModel(str, Enum):
    LLAMA3_70B = "llama3:70b"
    LLAMA3_8B = "llama3:8b"
    GEMMA2_9B = "gemma2:9b"
    QWEN2_7B = "qwen2:7b"
